"""Pipeline utils."""

from typing import Any
from typing import List
from typing import Optional
from typing import Sequence

from lightautoml.dataset.base import LAMLDataset


def map_pipeline_names(input_names: Sequence[str], output_names: Sequence[str]) -> List[Optional[str]]:
    """Pipelines create name in the way 'prefix__feature_name'.

    Multiple pipelines will create names
    in the way 'prefix1__prefix2__feature_name'.
    This function maps initial features names to outputs.
    Result may be not exact in some rare cases,
    but it's ok for real pipelines.

    Args:
        input_names: Initial feature names.
        output_names: Output feature names.

    Returns:
        Mapping between feature names.

    """
    # TODO: Add assert here
    mapped: List[Optional[str]] = [None] * len(output_names)
    s_in = set(input_names)

    for n, name in enumerate(output_names):
        splitted = name.split("__")

        for i in range(len(splitted)):
            name = "__".join(splitted[i:])
            if name in s_in:
                mapped[n] = name
                break

    assert None not in mapped, "Can not infer names. For feature selection purposes use simple pipeline (one-to-one)"

    return mapped


def get_columns_by_role(dataset: LAMLDataset, role_name: str, **kwargs: Any) -> List[str]:
    """Search for columns with specific role and attributes when building pipeline.

    Args:
        dataset: Dataset to search.
        role_name: Name of features role.
        **kwargs: Specific parameters values to search.
            Example: search for categories with OHE processing only.

    Returns:
        List of str features names.

    """
    features = []
    inv_roles = dataset.inverse_roles
    for role in inv_roles:
        if role.name == role_name:
            flg = True
            # TODO: maybe refactor
            for k in kwargs:
                try:
                    attr = getattr(role, k)
                except AttributeError:
                    flg = False
                    break
                if attr != kwargs[k]:
                    flg = False
                    break
            if flg:
                features.extend(inv_roles[role])

    return sorted(features)



"""Image features transformers."""

import logging
import os
import pickle

from copy import deepcopy
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import torch

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import NumericRole
from ..image.image import CreateImageFeatures
from ..image.image import DeepTimmImageEmbedder
from ..image.utils import pil_loader
from ..text.utils import get_textarr_hash
from ..text.utils import single_text_hash
from .base import LAMLTransformer


logger = logging.getLogger(__name__)

NumpyOrPandas = Union[NumpyDataset, PandasDataset]


def path_check(dataset: LAMLDataset):
    """Check if all passed vars are path.

    Args:
        dataset: LAMLDataset to check.

    Raises:
         AssertionError: If non-path features are present.

    """
    roles = dataset.roles
    features = dataset.features
    for f in features:
        assert roles[f].name == "Path", "Only path accepted in this transformer"


class ImageFeaturesTransformer(LAMLTransformer):
    """Simple image histogram."""

    _fit_checks = (path_check,)
    _transform_checks = ()
    _fname_prefix = "img_hist"

    def __init__(
        self,
        hist_size: int = 30,
        is_hsv: bool = True,
        n_jobs: int = 4,
        loader: Callable = pil_loader,
    ):
        """Create normalized color histogram for rgb or hsv image.

        Args:
            hist_size: Number of bins for each channel.
            is_hsv: Convert image to hsv.
            n_jobs: Number of threads for multiprocessing.
            loader: Callable for reading image from path.

        """
        self.hist_size = hist_size
        self.is_hsv = is_hsv
        self.n_jobs = n_jobs
        self.loader = loader

    @property
    def features(self) -> List[str]:
        """Features list.

        Returns:
            List of features names.

        """
        return self._features

    def fit(self, dataset: NumpyOrPandas):
        """Init hist class and create feature names.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            self.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)
        # set transformer features

        dataset = dataset.to_pandas()
        df = dataset.data

        feats = []
        self.dicts = {}
        for n, i in enumerate(df.columns):
            fg = CreateImageFeatures(self.hist_size, self.is_hsv, self.n_jobs, self.loader)
            features = list(
                np.char.array([self._fname_prefix + "_"]) + np.char.array(fg.fe.get_names()) + np.char.array(["__" + i])
            )
            self.dicts[i] = {"fg": fg, "feats": features}
            feats.extend(features)
        self._features = feats
        return self

    def transform(self, dataset: NumpyOrPandas) -> NumpyDataset:
        """Transform image dataset to color histograms.

        Args:
            dataset: Pandas or Numpy dataset of image paths.

        Returns:
            Dataset with encoded text.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # transform
        roles = NumericRole()
        outputs = []
        for n, i in enumerate(df.columns):
            new_arr = self.dicts[i]["fg"].transform(df[i].values)
            output = dataset.empty().to_numpy()
            output.set_data(new_arr, self.dicts[i]["feats"], roles)
            outputs.append(output)
        # create resulted
        return dataset.empty().to_numpy().concat(outputs)


class AutoCVWrap(LAMLTransformer):
    """Calculate image embeddings.

    Args:
        model: Name of effnet model.
        weights_path: Path to saved weights.
        cache_dir: Path to cache directory or None.
        subs: Subsample to fit transformer. If ``None`` - full data.
        device: Torch device.
        n_jobs: Number of threads for dataloader.
        random_state: Random state to take subsample and set torch seed.
        batch_size: Batch size for embedding model.
        verbose: Verbose data processing.

    """

    _fit_checks = (path_check,)
    _transform_checks = ()
    _fname_prefix = "emb_cv"
    _emb_name = ""

    @property
    def features(self) -> List[str]:
        """Features list.

        Returns:
            List of features names.

        """
        return self._features

    def __init__(
        self,
        model="efficientnet_b0.ra_in1k",
        weights_path: Optional[str] = None,
        cache_dir: str = "./cache_CV",
        subs: Optional[Any] = None,
        device: torch.device = torch.device("cuda:0"),
        n_jobs: int = 4,
        random_state: int = 42,
        batch_size: int = 128,
        verbose: bool = True,
    ):
        self.embed_model = model
        self.random_state = random_state
        self.subs = subs
        self.dicts = {}
        self.cache_dir = cache_dir

        self.transformer = DeepTimmImageEmbedder(
            device,
            n_jobs,
            random_state,
            model,
            weights_path,
            batch_size,
            verbose,
        )
        self._emb_name = "DI_" + single_text_hash(self.embed_model)
        self.emb_size = self.transformer.model.feature_shape

    def fit(self, dataset: NumpyOrPandas):
        """Fit chosen transformer and create feature names.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            self.

        """
        for check_func in self._fit_checks:
            check_func(dataset)

        if self.cache_dir is not None:
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir)
        # set transformer features

        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # fit
        if self.subs is not None and df.shape[0] >= self.subs:
            subs = df.sample(n=self.subs, random_state=self.random_state)
        else:
            subs = df

        names = []
        for n, i in enumerate(subs.columns):
            feats = [self._fname_prefix + "_" + self._emb_name + "_" + str(x) + "__" + i for x in range(self.emb_size)]
            self.dicts[i] = {
                "transformer": deepcopy(self.transformer.fit(subs[i])),
                "feats": feats,
            }
            names.extend(feats)

        self._features = names
        return self

    def transform(self, dataset: NumpyOrPandas) -> NumpyDataset:
        """Transform dataset to image embeddings.

        Args:
            dataset: Pandas or Numpy dataset of image paths.

        Returns:
            Numpy dataset with image embeddings.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # transform
        roles = NumericRole()
        outputs = []

        for n, conlumn_name in enumerate(df.columns):
            if self.cache_dir is not None:
                full_hash = get_textarr_hash(df[conlumn_name]) + get_textarr_hash(self.dicts[conlumn_name]["feats"])
                fname = os.path.join(self.cache_dir, full_hash + ".pkl")

                if os.path.exists(fname):
                    logger.info3(f"Load saved dataset for {conlumn_name}")

                    with open(fname, "rb") as f:
                        new_arr = pickle.load(f)

                else:
                    new_arr = self.dicts[conlumn_name]["transformer"].transform(df[conlumn_name])
                    with open(fname, "wb") as f:
                        pickle.dump(new_arr, f)
            else:
                new_arr = self.dicts[conlumn_name]["transformer"].transform(df[conlumn_name])

            output = dataset.empty().to_numpy()
            output.set_data(new_arr, self.dicts[conlumn_name]["feats"], roles)
            outputs.append(output)
            logger.info3(f"Feature {conlumn_name} transformed")
        # create resulted
        return dataset.empty().to_numpy().concat(outputs)



"""Validation utils."""

from typing import Callable
from typing import Optional
from typing import Union
from typing import cast

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from .base import DummyIterator
from .base import HoldoutIterator
from .base import TrainValidIterator
from .np_iterators import get_numpy_iterator


NpDataset = Union[CSRSparseDataset, NumpyDataset, PandasDataset]


def create_validation_iterator(
    train: LAMLDataset,
    valid: Optional[LAMLDataset] = None,
    n_folds: Optional[int] = None,
    cv_iter: Optional[Callable] = None,
) -> TrainValidIterator:
    """Creates train-validation iterator.

    If train is one of common datasets types (``PandasDataset``, ``NumpyDataset``, ``CSRSparseDataset``)
    the :func:`~lightautoml.validation.np_iterators.get_numpy_iterator` will be used.
    Else if validation dataset is defined, the holdout-iterator will be used.
    Else the dummy iterator will be used.

    Args:
        train: Dataset to train.
        valid: Optional dataset for validate.
        n_folds: maximum number of folds to iterate. If ``None`` - iterate through all folds.
        cv_iter: Takes dataset as input and return an iterator of indexes of train/valid for train dataset.

    Returns:
        New iterator.

    """
    if type(train) in [PandasDataset, NumpyDataset, CSRSparseDataset]:
        train = cast(NpDataset, train)
        valid = cast(NpDataset, valid)
        iterator = get_numpy_iterator(train, valid, n_folds, cv_iter)

    else:
        if valid is not None:
            iterator = HoldoutIterator(train, valid)
        else:
            iterator = DummyIterator(train)

    return iterator



"""Metrics and loss functions for scikit-learn models."""

import logging

from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np

from .base import Loss
from .base import fw_rmsle


logger = logging.getLogger(__name__)


_sk_loss_mapping = {"rmsle": ("mse", fw_rmsle, np.expm1)}

_sk_force_metric = {
    "rmsle": ("mse", None, None),
}


class SKLoss(Loss):
    """Loss used for scikit-learn.

    Args:
        loss: One of default loss function.
            Valid are: 'logloss', 'mse', 'crossentropy', 'rmsle'.
        loss_params: Additional loss parameters.
        fw_func: Forward transformation.
            Used for transformation of target and item weights.
        bw_func: backward transformation.
            Used for predict values transformation.

    """

    def __init__(
        self,
        loss: str,
        loss_params: Optional[Dict] = None,
        fw_func: Optional[Callable] = None,
        bw_func: Optional[Callable] = None,
    ):
        assert loss in [
            "logloss",
            "mse",
            "mae",
            "crossentropy",
            "rmsle",
        ], "Not supported in sklearn in general case."
        self.flg_regressor = loss in ["mse", "rmsle"]

        if loss in _sk_loss_mapping:
            self.loss, fw_func, bw_func = _sk_loss_mapping[loss]
        else:
            self.loss = loss
            # set forward and backward transformations
            if fw_func is not None:
                self._fw_func = fw_func
            if bw_func is not None:
                self._bw_func = bw_func

        self.loss_params = loss_params

    def set_callback_metric(
        self,
        metric: Union[str, Callable],
        greater_is_better: Optional[bool] = None,
        metric_params: Optional[Dict] = None,
        task_name: Optional[str] = None,
    ):
        """Callback metric setter.

        Uses default callback of parent class `Loss`.

        Args:
            metric: Callback metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.
            task_name: Name of task.

        """
        if self.loss in _sk_force_metric:
            metric, greater_is_better, metric_params = _sk_force_metric[self.loss]
            logger.info2("For sklearn {0} callback metric switched to {1}".format(self.loss, metric))

        super().set_callback_metric(metric, greater_is_better, metric_params, task_name)



"""Dimension reduction transformers."""

from typing import List
from typing import Optional
from typing import Union

import numpy as np

from sklearn.decomposition import PCA
from sklearn.decomposition import TruncatedSVD

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import NumericRole
from .base import LAMLTransformer


# type - something that can be converted to pandas dataset
NumpyTransformable = Union[NumpyDataset, PandasDataset]
NumpyCSR = Union[NumpyDataset, CSRSparseDataset]


# TODO: move all checks to the utils
def numeric_check(dataset: LAMLDataset):
    """Check if all passed vars are categories.

    Args:
        dataset: Dataset to check.

    Raises:
        AssertionError: If there is non number role.

    """
    roles = dataset.roles
    features = dataset.features
    for f in features:
        assert roles[f].name == "Numeric", "Only numbers accepted in this transformer"


# TODO: merge into one transformer
class PCATransformer(LAMLTransformer):
    """PCA.

    Args:
        subs: Subsample to fit algorithm. If None - full data.
        random_state: Random state to take subsample.
        n_components: Number of PCA components

    """

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "pca"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        subs: Optional[int] = None,
        random_state: int = 42,
        n_components: int = 500,
    ):
        self.subs = subs
        self.random_state = random_state
        self.n_components = n_components
        self._pca = PCA
        self.pca = None

    def fit(self, dataset: NumpyTransformable):
        """Fit algorithm on dataset.

        Args:
            dataset: Sparse or Numpy dataset of text features.

        Returns:
            Self.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)
        # set transformer features

        # convert to accepted dtype and get attributes
        dataset = dataset.to_numpy()
        data = dataset.data
        self.n_components = np.minimum(self.n_components, data.shape[1] - 1)
        self.pca = self._pca(n_components=self.n_components, random_state=self.random_state)
        self.pca.fit(data)

        orig_name = dataset.features[0].split("__")[-1]

        feats = (
            np.char.array([self._fname_prefix + "_"])
            + np.arange(self.n_components).astype(str)
            + np.char.array(["__" + orig_name])
        )

        self._features = list(feats)
        return self

    def transform(self, dataset: NumpyTransformable) -> NumpyDataset:
        """Transform input dataset to PCA representation.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Numpy dataset with text embeddings.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_numpy()
        data = dataset.data
        # transform
        data = self.pca.transform(data)

        # create resulted
        output = dataset.empty().to_numpy()
        output.set_data(data, self.features, NumericRole(np.float32))
        return output


class SVDTransformer(LAMLTransformer):
    """TruncatedSVD.

    Args:
        subs: Subsample to fit algorithm. If None - full data.
        random_state: Random state to take subsample.
        n_components: Number of SVD components.

    """

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "svd"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        subs: Optional[int] = None,
        random_state: int = 42,
        n_components: int = 100,
    ):
        self.subs = subs
        self.random_state = random_state
        self.n_components = n_components
        self._svd = TruncatedSVD
        self.svd = None

    def fit(self, dataset: NumpyCSR):
        """Fit algorithm on dataset.

        Args:
            dataset: Sparse or Numpy dataset of text features.

        Returns:
            self.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)
        # set transformer features

        # convert to accepted dtype and get attributes
        data = dataset.data
        self.n_components = np.minimum(self.n_components, data.shape[1] - 1)
        self.svd = self._svd(n_components=self.n_components, random_state=self.random_state)
        self.svd.fit(data)

        orig_name = dataset.features[0].split("__")[-1]

        feats = (
            np.char.array([self._fname_prefix + "_"])
            + np.arange(self.n_components).astype(str)
            + np.char.array(["__" + orig_name])
        )

        self._features = list(feats)
        return self

    def transform(self, dataset: NumpyCSR) -> NumpyDataset:
        """Transform input dataset to SVD representation.

        Args:
            dataset: Sparse or Numpy dataset of text features.

        Returns:
            Numpy dataset with text embeddings.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        data = dataset.data
        # transform
        data = self.svd.transform(data)

        # create resulted
        output = dataset.empty().to_numpy()
        output.set_data(data, self.features, NumericRole(np.float32))
        return output



"""Set of loss functions for different machine learning algorithms."""

from .base import _valid_str_metric_names
from .cb import CBLoss
from .lgb import LGBLoss
from .xgb import XGBLoss
from .sklearn import SKLoss
from .torch import TORCHLoss
from .torch import TorchLossWrapper


__all__ = [
    "XGBLoss",
    "LGBLoss",
    "TORCHLoss",
    "SKLoss",
    "CBLoss",
    "_valid_str_metric_names",
    "TorchLossWrapper",
]



"""Basic classes for validation iterators."""

from copy import copy
from typing import Any
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import cast

from lightautoml.dataset.base import LAMLDataset
from lightautoml.pipelines.features.base import FeaturesPipeline


# from ..pipelines.selection.base import SelectionPipeline

# TODO: SOLVE CYCLIC IMPORT PROBLEM!!! add Selectors typing

Dataset = TypeVar("Dataset", bound=LAMLDataset)
CustomIdxs = Iterable[Tuple[Sequence, Sequence]]


# add checks here
# check for same columns in dataset
class TrainValidIterator:
    """Abstract class to train/validation iteration.

    Train/valid iterator:
    should implement `__iter__` and `__next__` for using in ml_pipeline.


    Args:
        train: Train dataset.
        **kwargs: Key-word parameters.

    """

    @property
    def features(self):
        """Dataset features names.

        Returns:
            List of features names.

        """
        return self.train.features

    def __init__(self, train: Dataset, **kwargs: Any):
        self.train = train
        for k in kwargs:
            self.__dict__[k] = kwargs[k]

    def __iter__(self) -> Iterable:
        """Abstract method. Creates iterator."""
        raise NotImplementedError

    def __len__(self) -> Optional[int]:
        """Abstract method. Get length of dataset."""
        raise NotImplementedError

    def get_validation_data(self) -> LAMLDataset:
        """Abstract method. Get validation sample."""
        raise NotImplementedError

    def apply_feature_pipeline(self, features_pipeline: FeaturesPipeline) -> "TrainValidIterator":
        """Apply features pipeline on train data.

        Args:
            features_pipeline: Composite transformation of features.

        Returns:
            Copy of object with transformed features.

        """
        train_valid = copy(self)
        train_valid.train = features_pipeline.fit_transform(train_valid.train)
        return train_valid

    # TODO: add typing
    def apply_selector(self, selector) -> "TrainValidIterator":
        """Select features on train data.

        Check if selector is fitted.
        If not - fit and then perform selection.
        If fitted, check if it's ok to apply.

        Args:
            selector: Uses for feature selection.

        Returns:
            Dataset with selected features.

        """
        if not selector.is_fitted:
            selector.fit(self)
        train_valid = copy(self)
        train_valid.train = selector.select(train_valid.train)
        return train_valid

    def convert_to_holdout_iterator(self) -> "HoldoutIterator":
        """Abstract method. Convert iterator to HoldoutIterator."""
        raise NotImplementedError


class DummyIterator(TrainValidIterator):
    """Simple Iterator which use train data as validation."""

    def __init__(self, train: Dataset):
        """Create iterator. WARNING: validation on train.

        Args:
            train: Train dataset.

        """
        self.train = train

    def __len__(self) -> Optional[int]:
        """Get 1 len.

        Returns:
            '1'.

        """
        return 1

    def __iter__(self) -> List[Tuple[None, Dataset, Dataset]]:
        """Simple iterable object.

        Returns:
            Iterable object for dataset, where for validation also uses train.

        """
        return iter([(None, self.train, self.train)])

    def get_validation_data(self) -> Dataset:
        """Just get validation sample.

        Returns:
            Whole train dataset.

        """
        return self.train

    def convert_to_holdout_iterator(self) -> "HoldoutIterator":
        """Convert iterator to hold-out-iterator.

        Returns:
            iterator: Holdout iterator with ``'train == valid'``.

        """
        return HoldoutIterator(self.train, self.train)


class HoldoutIterator(TrainValidIterator):
    """Iterator for classic holdout - just predefined train and valid samples."""

    def __init__(self, train: LAMLDataset, valid: LAMLDataset):
        """Create iterator.

        Args:
            train: Dataset of train data.
            valid: Dataset of valid data.

        """
        self.train = train
        self.valid = valid

    def __len__(self) -> Optional[int]:
        """Get 1 len.

        Returns:
            1

        """
        return 1

    def __iter__(self) -> Iterable[Tuple[None, LAMLDataset, LAMLDataset]]:
        """Simple iterable object.

        Returns:
            Iterable object for train validation dataset.

        """
        return iter([(None, self.train, self.valid)])

    def get_validation_data(self) -> LAMLDataset:
        """Just get validation sample.

        Returns:
            Whole validation dataset.

        """
        return self.valid

    def apply_feature_pipeline(self, features_pipeline: FeaturesPipeline) -> "HoldoutIterator":
        """Inplace apply features pipeline to iterator components.

        Args:
            features_pipeline: Features pipeline to apply.

        Returns:
            New iterator.

        """
        train_valid = cast("HoldoutIterator", super().apply_feature_pipeline(features_pipeline))
        train_valid.valid = features_pipeline.transform(train_valid.valid)

        return train_valid

    def apply_selector(self, selector) -> "HoldoutIterator":
        """Same as for basic class, but also apply to validation.

        Args:
            selector: Uses for feature selection.

        Returns:
            New iterator.

        """
        train_valid = cast("HoldoutIterator", super().apply_selector(selector))
        train_valid.valid = selector.select(train_valid.valid)

        return train_valid

    def convert_to_holdout_iterator(self) -> "HoldoutIterator":
        """Do nothing, just return itself.

        Returns:
            self.

        """
        return self


class CustomIterator(TrainValidIterator):
    """Iterator that uses function to create folds indexes.

    Useful for example - classic timeseries splits.

    """

    def __init__(self, train: LAMLDataset, iterator: CustomIdxs):
        """Create iterator.

        Args:
            train: Dataset of train data.
            iterator: Callable(dataset) -> Iterator of train/valid indexes.

        """
        self.train = train
        self.iterator = iterator

    def __len__(self) -> Optional[int]:
        """Empty __len__ method.

        Returns:
            None.

        """
        return len(self.iterator)

    def __iter__(self) -> Generator:
        """Create generator of train/valid datasets.

        Returns:
            Data generator.

        """
        generator = ((val_idx, self.train[tr_idx], self.train[val_idx]) for (tr_idx, val_idx) in self.iterator)

        return generator

    def get_validation_data(self) -> LAMLDataset:
        """Simple return train dataset.

        Returns:
            Dataset of train data.

        """
        return self.train

    def convert_to_holdout_iterator(self) -> "HoldoutIterator":
        """Convert iterator to hold-out-iterator.

        Use first train/valid split for :class:`~lightautoml.validation.base.HoldoutIterator` creation.

        Returns:
            New hold out iterator.

        """
        for (tr_idx, val_idx) in self.iterator:
            return HoldoutIterator(self.train[tr_idx], self.train[val_idx])



"""Node utils models."""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.autograd import Function
from torch.jit import script


def check_numpy(x):
    """Makes sure x is a numpy array.

    Args:
        x : array to check.

    Returns:
        x
    """
    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    x = np.asarray(x)
    assert isinstance(x, np.ndarray)
    return x


def to_one_hot(y, depth=None):
    """Takes integer with n dims and converts it to 1-hot representation with n + 1 dims.

    The n+1'st dimension will have zeros everywhere but at y'th index, where it will be equal to 1.

    Args:
        y : input integer (IntTensor, LongTensor or Variable) of any shape
        depth : the size of the one hot dimension

    Returns:
        one hot Tensor
    """
    y_flat = y.to(torch.int64).view(-1, 1)
    depth = depth if depth is not None else int(torch.max(y_flat)) + 1
    y_one_hot = torch.zeros(y_flat.size()[0], depth, device=y.device).scatter_(1, y_flat, 1)
    y_one_hot = y_one_hot.view(*(tuple(y.shape) + (-1,)))
    return y_one_hot


def _make_ix_like(input, dim=0):
    d = input.size(dim)
    rho = torch.arange(1, d + 1, device=input.device, dtype=input.dtype)
    view = [1] * input.dim()
    view[0] = -1
    return rho.view(view).transpose(0, dim)


class SparsemaxFunction(Function):
    """An implementation of sparsemax (Martins & Astudillo, 2016).

    See :cite:`DBLP:journals/corr/MartinsA16` for detailed description.
    By Ben Peters and Vlad Niculae
    """

    @staticmethod
    def forward(ctx, input, dim=-1):
        """sparsemax: normalizing sparse transform (a la softmax).

        Args:
            ctx: context, to increase the speed
            input (Tensor): any shape
            dim: dimension along which to apply sparsemax

        Returns:
            Tensor same shape as input
        """
        ctx.dim = dim
        max_val, _ = input.max(dim=dim, keepdim=True)
        input -= max_val  # same numerical stability trick as for softmax
        tau, supp_size = SparsemaxFunction._threshold_and_support(input, dim=dim)
        output = torch.clamp(input - tau, min=0)
        ctx.save_for_backward(supp_size, output)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        """backward-pass.

        Args:
            ctx: context, to increase the speed
            grad_output: grad from the next layers

        Returns:
            grad output
        """
        supp_size, output = ctx.saved_tensors
        dim = ctx.dim
        grad_input = grad_output.clone()
        grad_input[output == 0] = 0

        v_hat = grad_input.sum(dim=dim) / supp_size.to(output.dtype).squeeze()
        v_hat = v_hat.unsqueeze(dim)
        grad_input = torch.where(output != 0, grad_input - v_hat, grad_input)
        return grad_input, None

    @staticmethod
    def _threshold_and_support(input, dim=-1):
        """Sparsemax building block compute the threshold.

        Args:
            input: any dimension
            dim: dimension along which to apply the sparsemax

        Returns:
            the threshold value
        """
        input_srt, _ = torch.sort(input, descending=True, dim=dim)
        input_cumsum = input_srt.cumsum(dim) - 1
        rhos = _make_ix_like(input, dim)
        support = rhos * input_srt > input_cumsum

        support_size = support.sum(dim=dim).unsqueeze(dim)
        tau = input_cumsum.gather(dim, support_size - 1)
        tau /= support_size.to(input.dtype)
        return tau, support_size


class Sparsemax(nn.Module):
    """Py-Torch class for Sparsemax."""

    def __init__(self):
        super(Sparsemax, self).__init__()

    def forward(self, input, dim=-1):
        """Forward-pass.

        Args:
            input (Tensor): input Tensor.
            dim (int): dimension which will be aggregatedю

        Returns:
            Entmax15(input,dim=dim)
        """
        return SparsemaxFunction.apply(input, dim)


class Sparsemoid(nn.Module):
    """Py-Torch class for Sparsemoid."""

    def __init__(self):
        super(Sparsemoid, self).__init__()

    def forward(self, input):
        """Forward-pass.

        Args:
            input (Tensor): input Tensor

        Returns:
            Sparsemoid(input)
        """
        return (0.5 * input + 0.5).clamp_(0, 1)


class Entmax15Function(Function):
    """An implementation of exact Entmax with alpha=1.5 (B. Peters, V. Niculae, A. Martins).

    See :cite:`https://arxiv.org/abs/1905.05702 for detailed description.
    Source: https://github.com/deep-spin/entmax
    """

    @staticmethod
    def forward(ctx, input, dim=-1):
        """Entmax: normalizing sparse transform (a la softmax).

        Args:
            ctx: context, to increase the speed
            input (Tensor): any shape
            dim: dimension along which to apply Entmax

        Returns:
            output (Tensor): same shape as input
        """
        ctx.dim = dim
        max_val, _ = input.max(dim=dim, keepdim=True)
        input = input - max_val  # same numerical stability trick as for softmax
        input = input / 2  # divide by 2 to solve actual Entmax

        tau_star, _ = Entmax15Function._threshold_and_support(input, dim)
        output = torch.clamp(input - tau_star, min=0) ** 2
        ctx.save_for_backward(output)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        """backward-pass.

        Args:
            ctx: context, to increase the speed
            grad_output: grad from the next layers

        Returns:
            grad output
        """
        (Y,) = ctx.saved_tensors
        gppr = Y.sqrt()  # = 1 / g'' (Y)
        dX = grad_output * gppr
        q = dX.sum(ctx.dim) / gppr.sum(ctx.dim)
        q = q.unsqueeze(ctx.dim)
        dX -= q * gppr
        return dX, None

    @staticmethod
    def _threshold_and_support(input, dim=-1):
        """Sparsemax building block compute the threshold.

        Args:
            input: any dimension
            dim: dimension along which to apply the sparsemax

        Returns:
            the threshold value
        """
        Xsrt, _ = torch.sort(input, descending=True, dim=dim)

        rho = _make_ix_like(input, dim)
        mean = Xsrt.cumsum(dim) / rho
        mean_sq = (Xsrt ** 2).cumsum(dim) / rho
        ss = rho * (mean_sq - mean ** 2)
        delta = (1 - ss) / rho

        # NOTE this is not exactly the same as in reference algo
        # Fortunately it seems the clamped values never wrongly
        # get selected by tau <= sorted_z. Prove this!
        delta_nz = torch.clamp(delta, 0)
        tau = mean - torch.sqrt(delta_nz)

        support_size = (tau <= Xsrt).sum(dim).unsqueeze(dim)
        tau_star = tau.gather(dim, support_size - 1)
        return tau_star, support_size


class Entmoid15Optimized(Function):
    """A highly optimized equivalent of lambda x: Entmax15([x, 0])."""

    @staticmethod
    def forward(ctx, input):
        """Entmoid15 (a la softmax).

        Args:
            ctx: context, to increase the speed
            input (Tensor): any shape

        Returns:
            output (Tensor): same shape as input
        """
        output = Entmoid15Optimized._forward(input)
        ctx.save_for_backward(output)
        return output

    @staticmethod
    @script
    def _forward(input):
        input, is_pos = abs(input), input >= 0
        tau = (input + torch.sqrt(F.relu(8 - input ** 2))) / 2
        tau.masked_fill_(tau <= input, 2.0)
        y_neg = 0.25 * F.relu(tau - input, inplace=True) ** 2
        return torch.where(is_pos, 1 - y_neg, y_neg)

    @staticmethod
    def backward(ctx, grad_output):
        """backward-pass.

        Args:
            ctx: context, to increase the speed
            grad_output: grad from the next layers

        Returns:
            grad output
        """
        return Entmoid15Optimized._backward(ctx.saved_tensors[0], grad_output)

    @staticmethod
    @script
    def _backward(output, grad_output):
        gppr0, gppr1 = output.sqrt(), (1 - output).sqrt()
        grad_input = grad_output * gppr0
        q = grad_input / (gppr0 + gppr1)
        grad_input -= q * gppr0
        return grad_input


class Entmax15(nn.Module):
    """Py-Torch class for Entmax15."""

    def __init__(self):
        super(Entmax15, self).__init__()

    def forward(self, input, dim=-1):
        """Forward-pass.

        Args:
            input (Tensor): input Tensor.
            dim (int): dimension which will be aggregatedю

        Returns:
            Entmax15(input,dim=dim)
        """
        return Entmax15Function.apply(input, dim)


class Entmoid15(nn.Module):
    """Py-Torch class for Entmoid15."""

    def __init__(self):
        super(Entmoid15, self).__init__()

    def forward(self, input):
        """Forward-pass.

        Args:
            input (Tensor): input Tensor

        Returns:
            Entmoid15(input)
        """
        return Entmoid15Optimized.apply(input)


class MeanPooling(nn.Module):
    """Pytorch implementation of MeanPooling head.

    Args:
        n_out: int, output dim.
        dim: int: the dimension to be averaged.
    """

    def __init__(self, n_out, dim=-1):
        super().__init__()
        self.n_out = n_out
        self.dim = dim

    def forward(self, x: torch.Tensor):
        """Forward-pass.

        # noqa: DAR101

        Returns:
            x[..., :self.n_out].mean(dim=self.dim)
        """
        return x[..., : self.n_out].mean(dim=self.dim)


class ModuleWithInit(nn.Module):
    """Base class for pytorch module with data-aware initializer on first batch."""

    def __init__(self):
        super().__init__()
        self._is_initialized_tensor = nn.Parameter(torch.tensor(0, dtype=torch.uint8), requires_grad=False)
        self._is_initialized_bool = None
        # Note: this module uses a separate flag self._is_initialized so as to achieve both
        # * persistence: is_initialized is saved alongside model in state_dict
        # * speed: model doesn't need to cache
        # please DO NOT use these flags in child modules

    def initialize(self, *args, **kwargs):
        """Initialize module tensors using first batch of data."""
        raise NotImplementedError("Please implement ")

    def __call__(self, *args, **kwargs):
        """Initialize module after forward-pass.

        # noqa: DAR101

        Returns:
            Forward-pass.
        """
        if self._is_initialized_bool is None:
            self._is_initialized_bool = bool(self._is_initialized_tensor.item())
        if not self._is_initialized_bool:
            self.initialize(*args, **kwargs)
            self._is_initialized_tensor.data[...] = 1
            self._is_initialized_bool = True
        return super().__call__(*args, **kwargs)


class ODST(ModuleWithInit):
    r"""Oblivious Differentiable Sparsemax Trees. http://tinyurl.com/odst-readmore.

    One can drop (sic!) this module anywhere instead of nn.Linear

    Args:
        in_features: number of features in the input tensor
        num_trees: number of trees in this layer
        tree_dim: number of response channels in the response of individual tree
        depth: number of splits in every tree
        flatten_output: if False, returns [..., num_trees, tree_dim],
            by default returns [..., num_trees * tree_dim]
        choice_function: f(tensor, dim) -> R_simplex computes feature weights s.t. f(tensor, dim).sum(dim) == 1
        bin_function: f(tensor) -> R[0, 1], computes tree leaf weights
        initialize_response_: in-place initializer for tree output tensor
        initialize_selection_logits_: in-place initializer for logits that select features for the tree
            both thresholds and scales are initialized with data-aware init (or .load_state_dict)
        threshold_init_beta: initializes threshold to a q-th quantile of data points
            where q ~ Beta(:threshold_init_beta:, :threshold_init_beta:)
            If this param is set to 1, initial thresholds will have the same distribution as data points
            If greater than 1 (e.g. 10), thresholds will be closer to median data value
            If less than 1 (e.g. 0.1), thresholds will approach min/max data values.
        threshold_init_cutoff: threshold log-temperatures initializer, \in (0, inf)
            By default(1.0), log-remperatures are initialized in such a way that all bin selectors
            end up in the linear region of sparse-sigmoid. The temperatures are then scaled by this parameter.
            Setting this value > 1.0 will result in some margin between data points and sparse-sigmoid cutoff value
            Setting this value < 1.0 will cause (1 - value) part of data points to end up in flat sparse-sigmoid region
            For instance, threshold_init_cutoff = 0.9 will set 10% points equal to 0.0 or 1.0
            Setting this value > 1.0 will result in a margin between data points and sparse-sigmoid cutoff value
            All points will be between (0.5 - 0.5 / threshold_init_cutoff) and (0.5 + 0.5 / threshold_init_cutoff)
    """

    def __init__(
        self,
        in_features,
        num_trees,
        depth=6,
        tree_dim=1,
        flatten_output=True,
        choice_function=Entmax15(),
        bin_function=Entmoid15(),
        initialize_response_=nn.init.normal_,
        initialize_selection_logits_=nn.init.uniform_,
        threshold_init_beta=1.0,
        threshold_init_cutoff=1.0,
    ):
        super().__init__()
        self.depth, self.num_trees, self.tree_dim, self.flatten_output = depth, num_trees, tree_dim, flatten_output
        self.choice_function, self.bin_function = choice_function, bin_function
        self.threshold_init_beta, self.threshold_init_cutoff = threshold_init_beta, threshold_init_cutoff

        self.response = nn.Parameter(torch.zeros([num_trees, tree_dim, 2 ** depth]), requires_grad=True)
        initialize_response_(self.response)

        self.feature_selection_logits = nn.Parameter(torch.zeros([in_features, num_trees, depth]), requires_grad=True)
        initialize_selection_logits_(self.feature_selection_logits)

        self.feature_thresholds = nn.Parameter(
            torch.full([num_trees, depth], float("nan"), dtype=torch.float32), requires_grad=True
        )  # nan values will be initialized on first batch (data-aware init)

        self.log_temperatures = nn.Parameter(
            torch.full([num_trees, depth], float("nan"), dtype=torch.float32), requires_grad=True
        )

        # binary codes for mapping between 1-hot vectors and bin indices
        with torch.no_grad():
            indices = torch.arange(2 ** self.depth)
            offsets = 2 ** torch.arange(self.depth)
            bin_codes = (indices.view(1, -1) // offsets.view(-1, 1) % 2).to(torch.float32)
            bin_codes_1hot = torch.stack([bin_codes, 1.0 - bin_codes], dim=-1)
            self.bin_codes_1hot = nn.Parameter(bin_codes_1hot, requires_grad=False)
            # ^-- [depth, 2 ** depth, 2]

    def forward(self, input):
        """Forward-pass.

        Args:
            input: any shape

        Returns:
            response
        """
        assert len(input.shape) >= 2
        if len(input.shape) > 2:
            return self.forward(input.view(-1, input.shape[-1])).view(*input.shape[:-1], -1)
        # new input shape: [batch_size, in_features]

        feature_logits = self.feature_selection_logits
        feature_selectors = self.choice_function(feature_logits, dim=0)
        # ^--[in_features, num_trees, depth]

        feature_values = torch.einsum("bi,ind->bnd", input, feature_selectors)
        # ^--[batch_size, num_trees, depth]

        threshold_logits = (feature_values - self.feature_thresholds) * torch.exp(-self.log_temperatures)

        threshold_logits = torch.stack([-threshold_logits, threshold_logits], dim=-1)
        # ^--[batch_size, num_trees, depth, 2]

        bins = self.bin_function(threshold_logits)
        # ^--[batch_size, num_trees, depth, 2], approximately binary

        bin_matches = torch.einsum("btds,dcs->btdc", bins, self.bin_codes_1hot)
        # ^--[batch_size, num_trees, depth, 2 ** depth]

        response_weights = torch.prod(bin_matches, dim=-2)
        # ^-- [batch_size, num_trees, 2 ** depth]

        response = torch.einsum("bnd,ncd->bnc", response_weights, self.response)
        # ^-- [batch_size, num_trees, tree_dim]

        return response.flatten(1, 2) if self.flatten_output else response

    def initialize(self, input, eps=1e-6):
        """Initialization.

        Args:
            input: any dimension
            eps: extra epsilon as a temperature
        """
        # data-aware initializer
        assert len(input.shape) == 2
        with torch.no_grad():
            feature_selectors = self.choice_function(self.feature_selection_logits, dim=0)
            # ^--[in_features, num_trees, depth]

            feature_values = torch.einsum("bi,ind->bnd", input, feature_selectors)
            # ^--[batch_size, num_trees, depth]

            # initialize thresholds: sample random percentiles of data
            percentiles_q = 100 * np.random.beta(
                self.threshold_init_beta, self.threshold_init_beta, size=[self.num_trees, self.depth]
            )
            self.feature_thresholds.data[...] = torch.as_tensor(
                list(map(np.percentile, check_numpy(feature_values.flatten(1, 2).t()), percentiles_q.flatten())),
                dtype=feature_values.dtype,
                device=feature_values.device,
            ).view(self.num_trees, self.depth)

            # init temperatures: make sure enough data points are in the linear region of sparse-sigmoid
            temperatures = np.percentile(
                check_numpy(abs(feature_values - self.feature_thresholds)),
                q=100 * min(1.0, self.threshold_init_cutoff),
                axis=0,
            )

            # if threshold_init_cutoff > 1, scale everything down by it
            temperatures /= max(1.0, self.threshold_init_cutoff)
            self.log_temperatures.data[...] = torch.log(torch.as_tensor(temperatures) + eps)

    def __repr__(self):
        return "{}(in_features={}, num_trees={}, depth={}, tree_dim={}, flatten_output={})".format(
            self.__class__.__name__,
            self.feature_selection_logits.shape[0],
            self.num_trees,
            self.depth,
            self.tree_dim,
            self.flatten_output,
        )


class DenseODSTBlock(nn.Sequential):
    """The DenseBlock from https://github.com/Qwicen.

    Args:
        sinput_dim: Input dim.
        layer_dim: num trees in one layer.
        num_layers: number of forests.
        tree_dim: number of response channels in the response of individual tree.
        max_features: maximum number of features per input
        depth: number of splits in every tree.
        input_dropout: Dropout rate forest layer.
        choice_function: str `entmax` or `sparsmax`.
        flatten_output: flatten output or not.
    """

    def __init__(
        self,
        input_dim,
        layer_dim,
        num_layers,
        tree_dim=1,
        max_features=None,
        input_dropout=0.0,
        choice_function="entmax",
        flatten_output=True,
        **kwargs
    ):
        layers = []
        ch_f = Sparsemax() if choice_function == "sparsmax" else Entmax15()
        bin_f = Sparsemoid() if choice_function == "sparsmax" else Entmoid15()
        for i in range(num_layers):
            oddt = ODST(
                input_dim,
                layer_dim,
                tree_dim=tree_dim,
                flatten_output=True,
                choice_function=ch_f,
                bin_function=bin_f,
                **kwargs
            )
            input_dim = min(input_dim + layer_dim * tree_dim, max_features or float("inf"))
            layers.append(oddt)

        super().__init__(*layers)
        self.num_layers, self.layer_dim, self.tree_dim = num_layers, layer_dim, tree_dim
        self.max_features, self.flatten_output = max_features, flatten_output
        self.input_dropout = input_dropout

    def forward(self, x):
        """Forward-pass."""
        initial_features = x.shape[-1]
        for layer in self:
            layer_inp = x
            if self.max_features is not None:
                tail_features = min(self.max_features, layer_inp.shape[-1]) - initial_features
                if tail_features != 0:
                    layer_inp = torch.cat([layer_inp[..., :initial_features], layer_inp[..., -tail_features:]], dim=-1)
            """
            Originally it was:
                if self.training and self.input_dropout:
                    layer_inp = F.dropout(layer_inp, self.input_dropout)
            """
            if self.input_dropout:
                layer_inp = F.dropout(layer_inp, self.input_dropout, self.training)
            h = layer(layer_inp)
            x = torch.cat([x, h], dim=-1)

        outputs = x[..., initial_features:]
        if not self.flatten_output:
            outputs = outputs.view(*outputs.shape[:-1], self.num_layers * self.layer_dim, self.tree_dim)
        return outputs



"""Basic classes for features generation."""

import logging

from copy import copy
from copy import deepcopy
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from pandas import DataFrame
from pandas import Series

from ...dataset.base import LAMLDataset
from ...dataset.np_pd_dataset import NumpyDataset
from ...dataset.np_pd_dataset import PandasDataset
from ...dataset.roles import CategoryRole
from ...dataset.roles import ColumnRole
from ...dataset.roles import NumericRole
from ...transformers.base import ChangeRoles
from ...transformers.base import ColumnsSelector
from ...transformers.base import ConvertDataset
from ...transformers.base import LAMLTransformer
from ...transformers.base import SequentialTransformer
from ...transformers.base import UnionTransformer
from ...transformers.categorical import CatIntersectstions
from ...transformers.categorical import FreqEncoder
from ...transformers.categorical import LabelEncoder
from ...transformers.categorical import MultiClassTargetEncoder
from ...transformers.categorical import MultioutputTargetEncoder
from ...transformers.categorical import OrdinalEncoder
from ...transformers.categorical import TargetEncoder
from ...transformers.datetime import BaseDiff
from ...transformers.datetime import DateSeasons
from ...transformers.groupby import GroupByTransformer
from ...transformers.numeric import FillInf
from ...transformers.numeric import FillnaMedian
from ...transformers.numeric import QuantileBinning
from ..utils import get_columns_by_role
from ..utils import map_pipeline_names


NumpyOrPandas = Union[PandasDataset, NumpyDataset]

logger = logging.getLogger(__name__)


class FeaturesPipeline:
    """Abstract class.

    Analyze train dataset and create composite transformer
    based on subset of features.
    Instance can be interpreted like Transformer
    (look for :class:`~lightautoml.transformers.base.LAMLTransformer`)
    with delayed initialization (based on dataset metadata)
    Main method, user should define in custom pipeline is ``.create_pipeline``.
    For example, look at
    :class:`~lightautoml.pipelines.features.lgb_pipeline.LGBSimpleFeatures`.
    After FeaturePipeline instance is created, it is used like transformer
    with ``.fit_transform`` and ``.transform`` method.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pipes: List[Callable[[LAMLDataset], LAMLTransformer]] = [self.create_pipeline]
        self.sequential = False

    # TODO: visualize pipeline ?
    @property
    def input_features(self) -> List[str]:
        """Names of input features of train data."""
        return self._input_features

    @input_features.setter
    def input_features(self, val: List[str]):
        """Setter for input_features.

        Args:
            val: List of strings.

        """
        self._input_features = deepcopy(val)

    @property
    def output_features(self) -> List[str]:
        """List of feature names that produces _pipeline."""
        return self._pipeline.features

    @property
    def used_features(self) -> List[str]:
        """List of feature names from original dataset that was used to produce output."""
        mapped = map_pipeline_names(self.input_features, self.output_features)
        return list(set(mapped))

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Analyse dataset and create composite transformer.

        Args:
            train: Dataset with train data.

        Returns:  # noqa DAR202
            Composite transformer (pipeline).

        """
        raise NotImplementedError

    def fit_transform(self, train: LAMLDataset) -> LAMLDataset:
        """Create pipeline and then fit on train data and then transform.

        Args:
            train: Dataset with train data.

        Returns:
            Dataset with new features.

        """
        # TODO: Think about input/output features attributes
        self._input_features = train.features
        self._pipeline = self._merge_seq(train) if self.sequential else self._merge(train)

        return self._pipeline.fit_transform(train)

    def transform(self, test: LAMLDataset) -> LAMLDataset:
        """Apply created pipeline to new data.

        Args:
            test: Dataset with test data.

        Returns:
            Dataset with new features.

        """
        return self._pipeline.transform(test)

    def set_sequential(self, val: bool = True):  # noqa D102
        self.sequential = val
        return self

    def append(self, pipeline):  # noqa D102
        if isinstance(pipeline, FeaturesPipeline):
            pipeline = [pipeline]

        for _pipeline in pipeline:
            self.pipes.extend(_pipeline.pipes)

        return self

    def prepend(self, pipeline):  # noqa D102
        if isinstance(pipeline, FeaturesPipeline):
            pipeline = [pipeline]

        for _pipeline in reversed(pipeline):
            self.pipes = _pipeline.pipes + self.pipes

        return self

    def pop(self, i: int = -1) -> Optional[Callable[[LAMLDataset], LAMLTransformer]]:  # noqa D102
        if len(self.pipes) > 1:
            return self.pipes.pop(i)

    def _merge(self, data: LAMLDataset) -> LAMLTransformer:
        pipes = []
        for pipe in self.pipes:
            pipes.append(pipe(data))

        return UnionTransformer(pipes) if len(pipes) > 1 else pipes[-1]

    def _merge_seq(self, data: LAMLDataset) -> LAMLTransformer:
        pipes = []
        for pipe in self.pipes:
            _pipe = pipe(data)
            data = _pipe.fit_transform(data)
            pipes.append(_pipe)

        return SequentialTransformer(pipes) if len(pipes) > 1 else pipes[-1]


class EmptyFeaturePipeline(FeaturesPipeline):
    """Dummy feature pipeline - ``.fit_transform`` and transform do nothing."""

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Create empty pipeline.

        Args:
            train: Dataset with train data.

        Returns:
            Composite transformer (pipeline), that do nothing.

        """
        return LAMLTransformer()


class TabularDataFeatures:
    """Helper class contains basic features transformations for tabular data.

    This method can de shared by all tabular feature pipelines,
    to simplify ``.create_automl`` definition.
    """

    def __init__(self, **kwargs: Any):
        """Set default parameters for tabular pipeline constructor.

        Args:
            **kwargs: Additional parameters.

        """
        self.multiclass_te_co = 3
        self.top_intersections = 5
        self.max_intersection_depth = 3
        self.subsample = 10000
        self.random_state = 42
        self.feats_imp = None
        self.ascending_by_cardinality = False

        self.max_bin_count = 10
        self.sparse_ohe = "auto"

        self.groupby_types = ["delta_median", "delta_mean", "min", "max", "std", "mode", "is_mode"]
        self.groupby_triplets = []
        self.groupby_top_based_on = "cardinality"
        self.groupby_top_categorical = 3
        self.groupby_top_numerical = 3

        for k in kwargs:
            self.__dict__[k] = kwargs[k]

    @staticmethod
    def get_cols_for_datetime(train: NumpyOrPandas) -> Tuple[List[str], List[str]]:
        """Get datetime columns to calculate features.

        Args:
            train: Dataset with train data.

        Returns:
            2 list of features names - base dates and common dates.

        """
        base_dates = get_columns_by_role(train, "Datetime", base_date=True)
        datetimes = get_columns_by_role(train, "Datetime", base_date=False) + get_columns_by_role(
            train, "Datetime", base_date=True, base_feats=True
        )

        return base_dates, datetimes

    def get_datetime_diffs(self, train: NumpyOrPandas) -> Optional[LAMLTransformer]:
        """Difference for all datetimes with base date.

        Args:
            train: Dataset with train data.

        Returns:
            Transformer or ``None`` if no required features.

        """
        base_dates, datetimes = self.get_cols_for_datetime(train)
        if len(datetimes) == 0 or len(base_dates) == 0:
            return

        dt_processing = SequentialTransformer(
            [
                ColumnsSelector(keys=list(set(datetimes + base_dates))),
                BaseDiff(base_names=base_dates, diff_names=datetimes),
            ]
        )
        return dt_processing

    def get_datetime_seasons(
        self, train: NumpyOrPandas, outp_role: Optional[ColumnRole] = None
    ) -> Optional[LAMLTransformer]:
        """Get season params from dates.

        Args:
            train: Dataset with train data.
            outp_role: Role associated with output features.

        Returns:
            Transformer or ``None`` if no required features.

        """
        _, datetimes = self.get_cols_for_datetime(train)
        for col in copy(datetimes):
            if len(train.roles[col].seasonality) == 0 and train.roles[col].country is None:
                datetimes.remove(col)

        if len(datetimes) == 0:
            return

        if outp_role is None:
            outp_role = NumericRole(np.float32)

        date_as_cat = SequentialTransformer(
            [
                ColumnsSelector(keys=datetimes),
                DateSeasons(outp_role),
            ]
        )
        return date_as_cat

    @staticmethod
    def get_numeric_data(
        train: NumpyOrPandas,
        feats_to_select: Optional[List[str]] = None,
        prob: Optional[bool] = None,
    ) -> Optional[LAMLTransformer]:
        """Select numeric features.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.
            prob: Probability flag.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            if prob is None:
                feats_to_select = get_columns_by_role(train, "Numeric")
            else:
                feats_to_select = get_columns_by_role(train, "Numeric", prob=prob)

        if len(feats_to_select) == 0:
            return

        num_processing = SequentialTransformer(
            [
                ColumnsSelector(keys=feats_to_select),
                ConvertDataset(dataset_type=NumpyDataset),
                ChangeRoles(NumericRole(np.float32)),
            ]
        )

        return num_processing

    @staticmethod
    def get_freq_encoding(
        train: NumpyOrPandas, feats_to_select: Optional[List[str]] = None
    ) -> Optional[LAMLTransformer]:
        """Get frequency encoding part.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = get_columns_by_role(train, "Category", encoding_type="freq")

        if len(feats_to_select) == 0:
            return

        cat_processing = SequentialTransformer(
            [
                ColumnsSelector(keys=feats_to_select),
                FreqEncoder(),
            ]
        )
        return cat_processing

    def get_ordinal_encoding(
        self, train: NumpyOrPandas, feats_to_select: Optional[List[str]] = None
    ) -> Optional[LAMLTransformer]:
        """Get order encoded part.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = get_columns_by_role(train, "Category", ordinal=True)

        if len(feats_to_select) == 0:
            return

        cat_processing = SequentialTransformer(
            [
                ColumnsSelector(keys=feats_to_select),
                OrdinalEncoder(subs=self.subsample, random_state=self.random_state),
            ]
        )
        return cat_processing

    def get_categorical_raw(
        self, train: NumpyOrPandas, feats_to_select: Optional[List[str]] = None
    ) -> Optional[LAMLTransformer]:
        """Get label encoded categories data.

        Args:
            train: Dataset with train data.
            feats_to_select: Features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = []
            for i in ["auto", "oof", "int", "ohe"]:
                feats_to_select.extend(get_columns_by_role(train, "Category", encoding_type=i))

        if len(feats_to_select) == 0:
            return

        cat_processing = [
            ColumnsSelector(keys=feats_to_select),
            LabelEncoder(subs=self.subsample, random_state=self.random_state),
        ]
        cat_processing = SequentialTransformer(cat_processing)
        return cat_processing

    def get_target_encoder(self, train: NumpyOrPandas) -> Optional[type]:
        """Get target encoder func for dataset.

        Args:
            train: Dataset with train data.

        Returns:
            Class

        """
        target_encoder = None
        if train.folds is not None:
            if train.task.name in ["binary", "reg"]:
                target_encoder = TargetEncoder
            elif (train.task.name == "multi:reg") or (train.task.name == "multilabel"):
                n_classes = train.target.shape[1]
                if n_classes <= self.multiclass_te_co:
                    target_encoder = MultioutputTargetEncoder

            else:
                n_classes = train.target.max() + 1
                if n_classes <= self.multiclass_te_co:
                    target_encoder = MultiClassTargetEncoder

        return target_encoder

    def get_binned_data(
        self, train: NumpyOrPandas, feats_to_select: Optional[List[str]] = None
    ) -> Optional[LAMLTransformer]:
        """Get encoded quantiles of numeric features.

        Args:
            train: Dataset with train data.
            feats_to_select: features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:
            feats_to_select = get_columns_by_role(train, "Numeric", discretization=True)

        if len(feats_to_select) == 0:
            return

        binned_processing = SequentialTransformer(
            [
                ColumnsSelector(keys=feats_to_select),
                QuantileBinning(nbins=self.max_bin_count),
            ]
        )
        return binned_processing

    def get_categorical_intersections(
        self, train: NumpyOrPandas, feats_to_select: Optional[List[str]] = None
    ) -> Optional[LAMLTransformer]:
        """Get transformer that implements categorical intersections.

        Args:
            train: Dataset with train data.
            feats_to_select: features to handle. If ``None`` - default filter.

        Returns:
            Transformer.

        """
        if feats_to_select is None:

            categories = get_columns_by_role(train, "Category")
            feats_to_select = categories

            if len(categories) <= 1:
                return

            if self.max_intersection_depth <= 1 or self.top_intersections <= 1:
                return
            elif len(categories) > self.top_intersections:
                feats_to_select = self.get_top_categories(train, mode="cat_intersections", top_n=self.top_intersections)

        elif len(feats_to_select) <= 1:
            return

        cat_processing = [
            ColumnsSelector(keys=feats_to_select),
            CatIntersectstions(
                subs=self.subsample,
                random_state=self.random_state,
                max_depth=self.max_intersection_depth,
            ),
        ]
        cat_processing = SequentialTransformer(cat_processing)

        return cat_processing

    def get_uniques_cnt(self, train: NumpyOrPandas, feats: List[str]) -> Series:
        """Get unique values cnt.

        Args:
            train: Dataset with train data.
            feats: Features names.

        Returns:
            Series.

        """
        uns = []
        for col in feats:
            feat = Series(train[:, col].data)
            if self.subsample is not None and self.subsample < len(feat):
                feat = feat.sample(
                    n=int(self.subsample) if self.subsample > 1 else None,
                    frac=self.subsample if self.subsample <= 1 else None,
                    random_state=self.random_state,
                )

            un = feat.value_counts(dropna=False)
            uns.append(un.shape[0])

        return Series(uns, index=feats, dtype="int")

    def get_top_categories(self, train: NumpyOrPandas, mode: str, top_n: int = 5) -> List[str]:
        """Get top categories by importance.

        If feature importance is not defined,
        or feats has same importance - sort it by unique values counts.
        In second case init param ``ascending_by_cardinality``
        defines how - asc or desc.

        Args:
            train: Dataset with train data.
            mode: What feature generation mode is used. Can be "cat_intersections" or "groupby".
            top_n: Number of top categories.

        Returns:
            List.

        """
        assert mode in ["cat_intersections", "groupby"]
        cats = get_columns_by_role(train, "Category")
        if len(cats) == 0 or top_n == 0:
            return []
        elif len(cats) <= top_n:
            return cats
        df = DataFrame({"importance": 0, "cardinality": 0}, index=cats)
        # importance if defined
        if self.feats_imp is not None:
            feats_imp = Series(self.feats_imp.get_features_score()).sort_values(ascending=False)
            df["importance"] = feats_imp[feats_imp.index.isin(cats)]
            df["importance"].fillna(-np.inf)
        # check for cardinality
        df["cardinality"] = self.get_uniques_cnt(train, cats)
        # sort
        if mode == "groupby" and self.groupby_top_based_on == "cardinality" or self.feats_imp is None:
            df = df.sort_values(by="cardinality", ascending=self.ascending_by_cardinality)
        else:
            df = df.sort_values(by="importance", ascending=False)
        # get top n
        top = list(df.index[:top_n])
        return top

    def get_top_numeric(self, train: NumpyOrPandas, top_n: int = 5) -> List[str]:
        """Get top numeric features by importance.

        If feature importance is not defined,
        or feats has same importance - sort it by unique values counts.
        In second case init param ``ascending_by_cardinality``
        defines how - asc or desc.

        Args:
            train: Dataset with train data.
            top_n: Number of top numeric features.

        Returns:
            List.
        """
        nums = get_columns_by_role(train, "Numeric")
        if len(nums) == 0 or top_n == 0:
            return []
        elif len(nums) <= top_n:
            return nums
        df = DataFrame({"importance": 0, "cardinality": 0}, index=nums)
        # importance if defined
        if self.feats_imp is not None:
            feats_imp = Series(self.feats_imp.get_features_score()).sort_values(ascending=False)
            df["importance"] = feats_imp[feats_imp.index.isin(nums)]
            df["importance"].fillna(-np.inf)
        # check for cardinality
        df["cardinality"] = -self.get_uniques_cnt(train, nums)
        # sort
        if self.groupby_top_based_on == "cardinality" or self.feats_imp is None:
            df = df.sort_values(by="cardinality", ascending=self.ascending_by_cardinality)
        else:
            df = df.sort_values(by="importance", ascending=False)
        # get top n
        top = list(df.index[:top_n])
        return top

    def get_groupby(self, train: NumpyOrPandas) -> Optional[LAMLTransformer]:
        """Get transformer that calculates group by features.

        Amount of features is limited to ``self.top_group_by_categorical`` and ``self.top_group_by_numerical`` fields.

        Args:
            train: Dataset with train data.

        Returns:
            Transformer.
        """
        categorical_names = get_columns_by_role(train, "Category")
        numerical_names = get_columns_by_role(train, "Numeric")

        groupby_transformations = []
        if len(self.groupby_triplets) > 0:
            for group_col, feat_name, trans in self.groupby_triplets:
                categorical_cols = [] if feat_name in numerical_names else [feat_name]
                numeric_cols = [] if feat_name in categorical_names else [feat_name]
                if len(categorical_cols) + len(numeric_cols) == 0:
                    logging.info2("Feature is incorrect or dropped by preselector: {}".format(feat_name))
                    continue
                if group_col not in categorical_names:
                    logging.info2("Groupby column is incorrect or dropped by preselector: {}".format(group_col))
                    continue
                new_transformation = {
                    "group_col": group_col,
                    "categorical_cols": categorical_cols,
                    "numeric_cols": numeric_cols,
                    "used_transforms": [trans],
                }
                groupby_transformations.append(new_transformation)
        else:
            cat_feats_to_select = self.get_top_categories(train, "groupby", self.groupby_top_categorical)
            num_feats_to_select = self.get_top_numeric(train, self.groupby_top_numerical)
            # At least two categoricals or one categorical and one numeric
            if len(cat_feats_to_select) < 1:
                return
            if len(cat_feats_to_select) == 1 and len(num_feats_to_select) < 1:
                return
            # collect groupby_transformations
            for i, group_col in enumerate(cat_feats_to_select):
                new_transformation = {
                    "group_col": group_col,
                    "categorical_cols": cat_feats_to_select[:i] + cat_feats_to_select[i + 1 :],
                    "numeric_cols": num_feats_to_select,
                    "used_transforms": self.groupby_types,
                }
                groupby_transformations.append(new_transformation)

        groupby_processing = [
            SequentialTransformer(
                [
                    UnionTransformer(
                        [
                            SequentialTransformer(
                                [
                                    ColumnsSelector(keys=[trans["group_col"]] + trans["categorical_cols"]),
                                    LabelEncoder(subs=None, random_state=42),
                                    ChangeRoles(CategoryRole(int)),
                                ]
                            ),
                            SequentialTransformer(
                                [ColumnsSelector(keys=trans["numeric_cols"]), FillInf(), FillnaMedian()]
                            ),
                        ]
                    ),
                    GroupByTransformer(**trans),
                ]
            )
            for trans in groupby_transformations
        ]
        groupby_processing = UnionTransformer(groupby_processing)

        return groupby_processing



"""Utility."""
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import torch

from lightautoml.automl.base import AutoML
from lightautoml.dataset.roles import ColumnRole
from lightautoml.dataset.roles import TargetRole
from lightautoml.dataset.roles import TreatmentRole
from lightautoml.ml_algo.linear_sklearn import LinearLBFGS
from lightautoml.pipelines.features.linear_pipeline import LinearFeatures
from lightautoml.pipelines.ml.base import MLPipeline
from lightautoml.reader.base import PandasToPandasReader
from lightautoml.tasks import Task


def create_linear_automl(
    task: Task,
    n_folds: int = 5,
    timeout: Optional[None] = None,
    n_reader_jobs: int = 1,
    cpu_limit: int = 4,
    # verbose: int = 0,
    random_state: int = 42,
):
    """Linear automl.

    Args:
        task: Task.
        n_folds: number of folds.
        timeout: Stub, not used.
        n_reader_jobs: Number of reader jobs.
        cpu_limit: CPU limit.
        random_state: random_state.

    Returns:
        automl:

    """
    torch.set_num_threads(cpu_limit)

    reader = PandasToPandasReader(
        task,
        cv=n_folds,
        random_state=random_state,
        n_jobs=n_reader_jobs,
    )
    pipe = LinearFeatures()
    model = LinearLBFGS()
    pipeline = MLPipeline(
        [model],
        pre_selection=None,
        features_pipeline=pipe,
        post_selection=None,
    )
    automl = AutoML(reader, [[pipeline]], skip_conn=False)  # , verbose=0)

    return automl


def _get_treatment_role(
    roles: Dict[
        Union[ColumnRole, str],
        Union[str, Sequence[str]],
    ]
) -> Tuple[Union[TreatmentRole, str], str]:
    """Extract treatment pair (key/val) from roles.

    Args:
        roles: Roles

    Returns:
        role, column name

    """
    treatment_role: Optional[Union[TreatmentRole, str]] = None
    treatment_col: str

    for k, v in roles.items():
        if isinstance(k, TreatmentRole) or (isinstance(k, str) and k == "treatment"):
            if not isinstance(v, str) and isinstance(v, Sequence):
                raise RuntimeError("Treatment column must be unique")
            else:
                treatment_role, treatment_col = k, v
                break

    if treatment_role is None:
        raise RuntimeError("Treatment role is absent")

    return treatment_role, treatment_col


def _get_target_role(
    roles: Dict[Union[ColumnRole, str], Union[str, Sequence[str]]]
) -> Tuple[Union[TargetRole, str], str]:
    """Extract target pair (key/val) from roles.

    Args:
        roles: Roles

    Returns:
        role, column name

    """
    target_role: Optional[Union[TargetRole, str]] = None
    target_col: str

    for k, v in roles.items():
        if isinstance(k, TargetRole) or (isinstance(k, str) and k == "target"):
            if isinstance(v, str):
                target_role, target_col = k, v
                break
            else:
                raise RuntimeError("Bad target column type")

    if target_role is None:
        raise RuntimeError("Target role is absent")

    return target_role, target_col



"""Datetime features transformers."""

from collections import OrderedDict
from typing import List
from typing import Optional
from typing import Sequence

import holidays
import numpy as np

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import CategoryRole
from ..dataset.roles import ColumnRole
from ..dataset.roles import NumericRole
from .base import LAMLTransformer


# type - dataset that is ok with datetime dtypes
DatetimeCompatible = PandasDataset

date_attrs = {
    "y": "year",
    "m": "month",
    "d": "day",
    "wd": "weekday",
    "doy": "dayofyear",
    "hour": "hour",
    "min": "minute",
    "sec": "second",
    "ms": "microsecond",
    "ns": "nanosecond",
}


def datetime_check(dataset: LAMLDataset):
    """Check if all passed vars are datetimes.

    Args:
        dataset: Dataset to check.

    Raises:
        AssertionError: If non-datetime features are present.

    """
    roles = dataset.roles
    features = dataset.features
    for f in features:
        assert roles[f].name == "Datetime", "Only datetimes accepted in this transformer"


class TimeToNum(LAMLTransformer):
    """Basic conversion strategy, used in selection one-to-one transformers.

    Datetime converted to difference with random date from the corresponding column.

    """

    basic_interval = "D"

    _fname_prefix = "dtdiff"
    _fit_checks = (datetime_check,)
    _transform_checks = ()

    def transform(self, dataset: DatetimeCompatible) -> NumpyDataset:
        """Transform dates to numeric differences with base date.

        Args:
            dataset: Numpy or Pandas dataset with datetime columns.

        Returns:
            Numpy dataset of numeric features.

        """
        # checks if exist
        super().transform(dataset)
        # convert to accepted format and get attributes
        dataset = dataset.to_pandas()
        data = dataset.data

        # transform
        roles = NumericRole(np.float32)

        new_arr = ((data - data.iloc[0]) / np.timedelta64(1, self.basic_interval)).values.astype(np.float32)
        # new_arr = data.apply(lambda x: ((x - x[0]) / np.timedelta64(1, self.basic_interval)), axis=1).values.astype(
        #     np.float32
        # )

        # create resulted
        output = dataset.empty().to_numpy()
        output.set_data(new_arr, self.features, roles)

        return output


class BaseDiff(LAMLTransformer):
    """Basic conversion strategy, used in selection one-to-one transformers.

    Datetime converted to difference with basic_date.

    Args:
        base_names: Base date names.
        diff_names: Difference date names.
        basic_interval: Time unit.

    """

    basic_interval = "D"

    _fname_prefix = "basediff"
    _fit_checks = (datetime_check,)
    _transform_checks = ()

    @property
    def features(self) -> List[str]:
        """List of features."""
        return self._features

    def __init__(
        self,
        base_names: Sequence[str],
        diff_names: Sequence[str],
        basic_interval: Optional[str] = "D",
    ):
        self.base_names = base_names
        self.diff_names = diff_names
        self.basic_interval = basic_interval

    def fit(self, dataset: LAMLDataset) -> "LAMLTransformer":
        """Fit transformer and return it's instance.

        Args:
            dataset: Dataset to fit on.

        Returns:
            self.

        """
        self._features = []
        for col in self.base_names:
            self._features.extend(["basediff_{0}__{1}".format(col, x) for x in self.diff_names])

        for check_func in self._fit_checks:
            check_func(dataset)
        return self

    def transform(self, dataset: DatetimeCompatible) -> NumpyDataset:
        """Transform dates to numeric differences with base date.

        Args:
            dataset: Numpy or Pandas dataset with datetime columns.

        Returns:
            NumpyDataset of numeric features.

        """
        # checks if exist
        super().transform(dataset)
        # convert to accepted format and get attributes
        dataset = dataset.to_pandas()
        data = dataset.data[self.diff_names].values
        base_cols = dataset.data[self.base_names]

        feats_block = []

        # transform
        for col in base_cols.columns:
            new_arr = ((data - base_cols[[col]].values) / np.timedelta64(1, self.basic_interval)).astype(np.float32)
            feats_block.append(new_arr)

        feats_block = np.concatenate(feats_block, axis=1)

        # create resulted
        output = dataset.empty().to_numpy()
        output.set_data(feats_block, self.features, NumericRole(dtype=np.float32))

        return output


class DateSeasons(LAMLTransformer):
    """Basic conversion strategy, used in selection one-to-one transformers.

    Datetime converted to difference with basic_date.

    Args:
        output_role: Which role to assign for input features.

    """

    _fname_prefix = "season"
    _fit_checks = (datetime_check,)
    _transform_checks = ()

    @property
    def features(self) -> List[str]:
        """List of features names."""
        return self._features

    def __init__(self, output_role: Optional[ColumnRole] = None):
        self.output_role = output_role
        if output_role is None:
            self.output_role = CategoryRole(np.int32)

    def fit(self, dataset: LAMLDataset) -> "LAMLTransformer":
        """Fit transformer and return it's instance.

        Args:
            dataset: LAMLDataset to fit on.

        Returns:
            self.

        """
        for check_func in self._fit_checks:
            check_func(dataset)

        feats = dataset.features
        roles = dataset.roles
        self._features = []
        self.transformations = OrderedDict()

        for col in feats:
            seas = roles[col].seasonality
            self.transformations[col] = seas
            for s in seas:
                self._features.append("season_{0}__{1}".format(s, col))
            if roles[col].country is not None:
                self._features.append("season_hol__{0}".format(col))

        return self

    def transform(self, dataset: DatetimeCompatible) -> NumpyDataset:
        """Transform dates to categories - seasons and holiday flag.

        Args:
            dataset: Numpy or Pandas dataset with datetime columns.

        Returns:
            Numpy dataset of numeric features.

        """
        # checks if exist
        super().transform(dataset)
        # convert to accepted format and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data
        roles = dataset.roles

        new_arr = np.empty((df.shape[0], len(self._features)), np.int32)

        n = 0
        for col in dataset.features:
            for seas in self.transformations[col]:
                new_arr[:, n] = getattr(df[col].dt, date_attrs[seas])
                n += 1

            if roles[col].country is not None:
                # get years
                years = np.unique(df[col].dt.year)
                hol = holidays.CountryHoliday(
                    roles[col].country,
                    years=years,
                    prov=roles[col].prov,
                    state=roles[col].state,
                )
                new_arr[:, n] = df[col].dt.date.isin(hol)
                n += 1

        # create resulted
        output = dataset.empty().to_numpy()
        output.set_data(new_arr, self.features, self.output_role)

        return output



"""Bunch of metrics with unified interface."""

from functools import partial
from typing import Callable
from typing import Optional

import numpy as np

from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
from sklearn.metrics import log_loss
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_squared_log_error
from sklearn.metrics import r2_score
from sklearn.metrics import roc_auc_score


def mean_quantile_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    q: float = 0.9,
) -> float:
    """Computes Mean Quantile Error.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Specify weighted mean.
        q: Metric coefficient.

    Returns:
        metric value.

    """
    err = y_pred - y_true
    s = np.sign(err)
    err = np.abs(err)
    err = np.where(s > 0, q, 1 - q) * err
    if sample_weight is not None:
        return (err * sample_weight).mean() / sample_weight.mean()

    return err.mean()


def mean_huber_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    a: float = 0.9,
) -> float:
    """Computes Mean Huber Error.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Specify weighted mean.
        a: Metric coefficient.

    Returns:
        Metric value.

    """
    err = y_pred - y_true
    s = np.abs(err) < a
    err = np.where(s, 0.5 * (err ** 2), a * np.abs(err) - 0.5 * (a ** 2))

    if sample_weight is not None:
        return (err * sample_weight).mean() / sample_weight.mean()

    return err.mean()


def mean_fair_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    c: float = 0.9,
) -> float:
    """Computes Mean Fair Error.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Specify weighted mean.
        c: Metric coefficient.

    Returns:
        Metric value.

    """
    x = np.abs(y_pred - y_true) / c
    err = c ** 2 * (x - np.log(x + 1))

    if sample_weight is not None:
        return (err * sample_weight).mean() / sample_weight.mean()

    return err.mean()


def mean_absolute_percentage_error(
    y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None
) -> float:
    """Computes Mean Absolute Percentage error.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Specify weighted mean.

    Returns:
        Metric value.

    """
    err = (y_true - y_pred) / y_true
    err = np.abs(err)

    if sample_weight is not None:
        return (err * sample_weight).mean() / sample_weight.mean()

    return err.mean()


def roc_auc_ovr(y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None):
    """ROC-AUC One-Versus-Rest.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Weights of samples.

    Returns:
        Metric values.

    """
    return roc_auc_score(y_true, y_pred, sample_weight=sample_weight, multi_class="ovr")


def rmsle(y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None):
    """Root mean squared log error.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Weights of samples.

    Returns:
        Metric values.


    """
    return np.sqrt(mean_squared_log_error(y_true, y_pred, sample_weight=sample_weight))


def auc_mu(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    class_weights: Optional[np.ndarray] = None,
) -> float:
    """Compute multi-class metric AUC-Mu.

    We assume that confusion matrix full of ones, except diagonal elements.
    All diagonal elements are zeroes.
    By default, for averaging between classes scores we use simple mean.

    Args:
        y_true: True target values.
        y_pred: Predicted target values.
        sample_weight: Not used.
        class_weights: The between classes weight matrix. If ``None``,
            the standard mean will be used. It is expected to be a lower
            triangular matrix (diagonal is also full of zeroes).
            In position (i, j), i > j, there is a partial positive score
            between i-th and j-th classes. All elements must sum up to 1.

    Returns:
        Metric value.

    Note:
        Code was refactored from https://github.com/kleimanr/auc_mu/blob/master/auc_mu.py

    """
    if not isinstance(y_pred, np.ndarray):
        raise TypeError("Expected y_pred to be np.ndarray, got: {}".format(type(y_pred)))
    if not y_pred.ndim == 2:
        raise ValueError("Expected array with predictions be a 2-dimentional array")
    if not isinstance(y_true, np.ndarray):
        raise TypeError("Expected y_true to be np.ndarray, got: {}".format(type(y_true)))
    if not y_true.ndim == 1:
        raise ValueError("Expected array with ground truths be a 1-dimentional array")
    if y_true.shape[0] != y_pred.shape[0]:
        raise ValueError(
            "Expected number of samples in y_true and y_pred be same,"
            " got {} and {}, respectively".format(y_true.shape[0], y_pred.shape[0])
        )

    uniq_labels = np.unique(y_true)
    n_samples, n_classes = y_pred.shape

    if not np.all(uniq_labels == np.arange(n_classes)):
        raise ValueError("Expected classes encoded values 0, ..., N_classes-1")

    if class_weights is None:
        class_weights = np.tri(n_classes, k=-1)
        class_weights /= class_weights.sum()

    if not isinstance(class_weights, np.ndarray):
        raise TypeError("Expected class_weights to be np.ndarray, got: {}".format(type(class_weights)))
    if not class_weights.ndim == 2:
        raise ValueError("Expected class_weights to be a 2-dimentional array")
    if not class_weights.shape == (n_classes, n_classes):
        raise ValueError("Expected class_weights size: {}, got: {}".format((n_classes, n_classes), class_weights.shape))
    # check sum?
    confusion_matrix = np.ones((n_classes, n_classes)) - np.eye(n_classes)
    auc_full = 0.0

    for class_i in range(n_classes):
        preds_i = y_pred[y_true == class_i]
        n_i = preds_i.shape[0]
        for class_j in range(class_i):
            preds_j = y_pred[y_true == class_j]
            n_j = preds_j.shape[0]
            n = n_i + n_j
            tmp_labels = np.zeros((n,), dtype=np.int32)
            tmp_labels[n_i:] = 1
            tmp_pres = np.vstack((preds_i, preds_j))
            v = confusion_matrix[class_i, :] - confusion_matrix[class_j, :]
            scores = np.dot(tmp_pres, v)
            score_ij = roc_auc_score(tmp_labels, scores)
            auc_full += class_weights[class_i, class_j] * score_ij

    return auc_full


class F1Factory:
    """Wrapper for :func:`~sklearn.metrics.f1_score` function.

    Args:
        average: Averaging type ('micro', 'macro', 'weighted').

    """

    def __init__(self, average: str = "micro"):
        self.average = average

    def __call__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        """Compute metric.

        Args:
            y_true: Ground truth target values.
            y_pred: Estimated target values.
            sample_weight: Sample weights.

        Returns:
            F1 score of the positive class in binary classification
            or weighted average of the F1 scores of each class
            for the multiclass task.

        """
        return f1_score(y_true, y_pred, sample_weight=sample_weight, average=self.average)


class BestClassBinaryWrapper:
    r"""Metric wrapper to get best class prediction instead of probs.

    There is cut-off for prediction by ``0.5``.

    Args:
        func: Metric function. Function format:
            func(y_pred, y_true, weights, \*\*kwargs).

    """

    def __init__(self, func: Callable):
        self.func = func

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None, **kwargs):
        """Calculate metric."""
        y_pred = (y_pred > 0.5).astype(np.float32)

        return self.func(y_true, y_pred, sample_weight=sample_weight, **kwargs)


class BestClassMulticlassWrapper:
    r"""Metric wrapper to get best class prediction instead of probs for multiclass.

    Prediction provides by argmax.

    Args:
        func: Metric function. Function format:
            func(y_pred, y_true, weights, \*\*kwargs)

    """

    def __init__(self, func):
        self.func = func

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None, **kwargs):
        """Calculate metric."""
        y_pred = (y_pred.argmax(axis=1)).astype(np.float32)

        return self.func(y_true, y_pred, sample_weight=sample_weight, **kwargs)


# TODO: Add custom metrics - precision/recall/fscore at K. Fscore at best co
# TODO: Move to other module


_valid_str_binary_metric_names = {
    "auc": roc_auc_score,
    "logloss": partial(log_loss),
    "accuracy": BestClassBinaryWrapper(accuracy_score),
}

_valid_str_reg_metric_names = {
    "r2": r2_score,
    "mse": mean_squared_error,
    "mae": mean_absolute_error,
    "rmsle": rmsle,
    "fair": mean_fair_error,
    "huber": mean_huber_error,
    "quantile": mean_quantile_error,
    "mape": mean_absolute_percentage_error,
}

_valid_str_multiclass_metric_names = {
    "auc_mu": auc_mu,
    "auc": roc_auc_ovr,
    "crossentropy": partial(log_loss),
    "accuracy": BestClassMulticlassWrapper(accuracy_score),
    "f1_macro": BestClassMulticlassWrapper(F1Factory("macro")),
    "f1_micro": BestClassMulticlassWrapper(F1Factory("micro")),
    "f1_weighted": BestClassMulticlassWrapper(F1Factory("weighted")),
}
_valid_str_multireg_metric_names = {"mse": mean_squared_error, "mae": mean_absolute_error}

_valid_str_multilabel_metric_names = {"logloss": partial(log_loss)}

_valid_str_metric_names = {
    "binary": _valid_str_binary_metric_names,
    "reg": _valid_str_reg_metric_names,
    "multiclass": _valid_str_multiclass_metric_names,
    "multi:reg": _valid_str_multireg_metric_names,
    "multilabel": _valid_str_multilabel_metric_names,
}

_valid_metric_args = {"quantile": ["q"], "huber": ["a"], "fair": ["c"]}



"""Contain Task object and metric wrappers."""

import inspect
import logging

from functools import partial
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np

from .common_metric import _valid_metric_args
from .common_metric import _valid_str_metric_names
from .losses import CBLoss
from .losses import LGBLoss
from .losses import XGBLoss
from .losses import SKLoss
from .losses import TORCHLoss
from .utils import infer_gib
from .utils import infer_gib_multiclass


if TYPE_CHECKING:
    from ..dataset.base import LAMLDataset
    from ..dataset.np_pd_dataset import NumpyDataset
    from ..dataset.np_pd_dataset import PandasDataset

    SklearnCompatible = Union[NumpyDataset, PandasDataset]

logger = logging.getLogger(__name__)

_valid_task_names = ["binary", "reg", "multiclass", "multi:reg", "multilabel"]
_one_dim_output_tasks = ["binary", "reg"]

_default_losses = {
    "binary": "logloss",
    "reg": "mse",
    "multiclass": "crossentropy",
    "multi:reg": "mae",
    "multilabel": "logloss",
}

_default_metrics = {
    "binary": "auc",
    "reg": "mse",
    "multiclass": "crossentropy",
    "multi:reg": "mae",
    "multilabel": "logloss",
}

_valid_loss_types = ["lgb", "sklearn", "torch", "cb"]

_valid_str_loss_names = {
    "binary": ["logloss"],
    "reg": ["mse", "mae", "mape", "rmsle", "quantile", "huber", "fair"],
    "multiclass": ["crossentropy", "f1"],
    "multi:reg": ["mae", "mse"],
    "multilabel": ["logloss"],
}


_valid_loss_args = {"quantile": ["q"], "huber": ["a"], "fair": ["c"]}


class LAMLMetric:
    """Abstract class for metric.

    Metric should be called on dataset.

    """

    greater_is_better = True

    def __call__(self, dataset: "LAMLDataset", dropna: bool = False):
        """Call metric on dataset.

        Args:
            dataset: Table with data.
            dropna: To ignore NaN in metric calculation.

        Returns:  # noqa DAR202
            Metric value.

        Raises:
            AttributeError: If metric isn't defined.

        """
        assert hasattr(dataset, "target"), "Dataset should have target to calculate metric"
        raise NotImplementedError


class ArgsWrapper:
    """Wrapper - ignore sample_weight if metric not accepts.

    Args:
        func: Metric function.
        metric_params: Additional metric parameters.

    """

    def __init__(self, func: Callable, metric_params: dict):
        keys = inspect.signature(func).parameters
        self.flg = "sample_weight" in keys
        self.func = partial(func, **metric_params)

    def __call__(self, y_true, y_pred, sample_weight=None):
        """Calculate metric value.

        If the metric does not include weights, then they are ignored.

        Args:
            y_true: Ground truth target values.
            y_pred: Estimated target values.
            sample_weight: Sample weights.

        Returns:
            Metric value.

        """
        if self.flg:
            return self.func(y_true, y_pred, sample_weight=sample_weight)

        return self.func(y_true, y_pred)


class SkMetric(LAMLMetric):
    """Abstract class for scikit-learn compatible metric.

    Implements metric calculation in sklearn format on numpy/pandas datasets.


    Args:
        metric: Specifies metric. Format:
            ``func(y_true, y_false, Optional[sample_weight], **kwargs)`` -> `float`.
        name: Name of metric.
        greater_is_better: Whether or not higher metric value is better.
        one_dim: `True` for single class, False for multiclass.
        weighted: Weights of classes.
        **kwargs: Other parameters for metric.

    """

    @property
    def metric(self) -> Callable:
        """Metric function."""
        assert self._metric is not None, "Metric calculation is not defined"
        return self._metric

    @property
    def name(self) -> str:
        """Name of used metric."""
        if self._name is None:
            return "AutoML Metric"
        else:
            return self._name

    def __init__(
        self,
        metric: Optional[Callable] = None,
        name: Optional[str] = None,
        greater_is_better: bool = True,
        one_dim: bool = True,
        **kwargs: Any,
    ):
        self._metric = metric
        self._name = name

        self.greater_is_better = greater_is_better
        self.one_dim = one_dim
        # self.weighted = weighted

        self.kwargs = kwargs

    def __call__(self, dataset: "SklearnCompatible", dropna: bool = False) -> float:
        """Implement call sklearn metric on dataset.

        Args:
            dataset: Dataset in Numpy or Pandas format.
            dropna: To ignore NaN in metric calculation.

        Returns:
            Metric value.

        Raises:
            AssertionError: if dataset has no target or
                target specified as one-dimensioned, but it is not.

        """
        assert hasattr(dataset, "target"), "Dataset should have target to calculate metric"
        if self.one_dim:
            assert dataset.shape[1] == 1, "Dataset should have single column if metric is one_dim"
        # TODO: maybe refactor this part?
        dataset = dataset.to_numpy()
        y_true = dataset.target
        y_pred = dataset.data
        sample_weight = dataset.weights

        if dropna:
            sl = ~np.isnan(y_pred).any(axis=1)
            y_pred = y_pred[sl]
            y_true = y_true[sl]
            if sample_weight is not None:
                sample_weight = sample_weight[sl]

        if self.one_dim:
            y_pred = y_pred[:, 0]

        value = self.metric(y_true, y_pred, sample_weight=sample_weight)
        sign = 2 * float(self.greater_is_better) - 1
        return value * sign


class Task:
    """Specify task (binary classification, multiclass classification, regression), metrics, losses.

    Args:
        name: Task name.
        loss: Objective function or dict of functions.
        loss_params: Additional loss parameters,
            if dict there is no presence check for loss_params.
        metric: String name or callable.
        metric_params: Additional metric parameters.
        greater_is_better: Whether or not higher value is better.

    Note:
        There is 3 different task types:

            - `'binary'` - for binary classification.
            - `'reg'` - for regression.
            - `'multiclass'` - for multiclass classification.

        Available losses for binary task:

            - `'logloss'` - (uses by default) Standard logistic loss.

        Available losses for regression task:

            - `'mse'` - (uses by default) Mean Squared Error.
            - `'mae'` - Mean Absolute Error.
            - `'mape'` - Mean Absolute Percentage Error.
            - `'rmsle'` - Root Mean Squared Log Error.
            - `'huber'` - Huber loss, required params:
                ``a`` - threshold between MAE and MSE losses.
            - `'fair'` - Fair loss, required params:
                ``c`` - sets smoothness.
            - `'quantile'` - Quantile loss, required params:
                ``q`` - sets quantile.

        Available losses for multi-classification task:

            - `'crossentropy'` - (uses by default) Standard crossentropy function.
            - `'f1'` - Optimizes F1-Macro Score, now available for
                LightGBM and NN models. Here we implicitly assume
                that the prediction lies not in the set ``{0, 1}``,
                but in the interval ``[0, 1]``.

        Available metrics for binary task:

            - `'auc'` - (uses by default) ROC-AUC score.
            - `'accuracy'` - Accuracy score (uses argmax prediction).
            - `'logloss'` - Standard logistic loss.

        Available metrics for regression task:

            - `'mse'` - (uses by default) Mean Squared Error.
            - `'mae'` - Mean Absolute Error.
            - `'mape'` - Mean Absolute Percentage Error.
            - `'rmsle'` - Root Mean Squared Log Error.
            - `'huber'` - Huber loss, required params:
                ``a`` - threshold between MAE and MSE losses.
            - `'fair'` - Fair loss, required params:
                ``c`` - sets smoothness.
            - `'quantile'` - Quantile loss, required params:
                ``q`` - sets quantile.

        Available metrics for multi-classification task:

            - `'crossentropy'` - (uses by default) Standard cross-entropy loss.
            - `'auc'` - ROC-AUC of each class against the rest.
            - `'auc_mu'` - AUC-Mu. Multi-class extension of standard AUC
                for binary classification. In short,
                mean of n_classes * (n_classes - 1) / 2 binary AUCs.
                More info on http://proceedings.mlr.press/v97/kleiman19a/kleiman19a.pdf


    Example:
        >>> task = Task('binary', metric='auc')

    """

    @property
    def name(self) -> str:
        """Name of task."""
        return self._name

    def __init__(
        self,
        name: str,
        loss: Optional[Union[dict, str]] = None,
        loss_params: Optional[Dict] = None,
        metric: Optional[Union[str, Callable]] = None,
        metric_params: Optional[Dict] = None,
        greater_is_better: Optional[bool] = None,
    ):

        assert name in _valid_task_names, "Invalid task name: {}, allowed task names: {}".format(
            name, _valid_task_names
        )
        self._name = name

        # add losses
        # if None - infer from task
        self.losses = {}
        if loss is None:
            loss = _default_losses[self.name]

        if loss_params is None:
            loss_params = {}

        # case - infer from string
        if type(loss) is str:

            # case when parameters defined
            if len(loss_params) > 0:
                self._check_loss_from_params(loss, loss_params)
                # check if loss and metric are the same - rewrite loss params
                # ??? "rewrite METRIC params" ???
                if loss == metric:
                    metric_params = loss_params
                    logger.info2("As loss and metric are equal, metric params are ignored.")

            else:
                assert (
                    loss not in _valid_loss_args
                ), "Loss should be defined with arguments. Ex. loss='quantile', loss_params={'q': 0.7}."
                loss_params = None

            assert loss in _valid_str_loss_names[self.name], "Invalid loss name: {} for task {}.".format(
                loss, self.name
            )

            for loss_key, loss_factory in zip(
                ["lgb", "sklearn", "torch", "cb", "xgb"], [LGBLoss, SKLoss, TORCHLoss, CBLoss, XGBLoss]
            ):
                try:
                    self.losses[loss_key] = loss_factory(loss, loss_params=loss_params)
                except (AssertionError, TypeError, ValueError):
                    logger.info2("{0} doesn't support in general case {1} and will not be used.".format(loss_key, loss))

                # self.losses[loss_key] = loss_factory(loss, loss_params=loss_params)

            assert len(self.losses) > 0, "None of frameworks supports {0} loss.".format(loss)

        elif type(loss) is dict:
            # case - dict passed directly
            # TODO: check loss parameters?
            #  Or it there will be assert when use functools.partial
            # assert all(map(lambda x: x in _valid_loss_types, loss)), 'Invalid loss key.'
            assert len([key for key in loss.keys() if key in _valid_loss_types]) != len(loss), "Invalid loss key."
            self.losses = loss

        else:
            raise TypeError("Loss passed incorrectly.")

        # set callback metric for loss
        # if no metric - infer from task
        if metric is None:
            metric = _default_metrics[self.name]

        self.metric_params = {}
        if metric_params is not None:
            self.metric_params = metric_params

        if type(metric) is str:

            self._check_metric_from_params(metric, self.metric_params)
            metric_func = _valid_str_metric_names[self.name][metric]
            metric_func = partial(metric_func, **self.metric_params)
            self.metric_func = metric_func
            self.metric_name = metric

        else:
            metric = ArgsWrapper(metric, self.metric_params)
            self.metric_params = {}
            self.metric_func = metric
            self.metric_name = None

        if greater_is_better is None:
            infer_gib_fn = infer_gib_multiclass if (name == "multiclass" or name == "multilabel") else infer_gib
            greater_is_better = infer_gib_fn(self.metric_func)

        self.greater_is_better = greater_is_better

        for loss_key in self.losses:
            try:
                self.losses[loss_key].set_callback_metric(metric, greater_is_better, self.metric_params, self.name)
            except:
                print(f"{self.name} isn`t supported in {loss_key}")

    def get_dataset_metric(self) -> LAMLMetric:
        """Create metric for dataset.

        Get metric that is called on dataset.

        Returns:
            Metric in scikit-learn compatible format.

        """
        # for now - case of sklearn metric only
        one_dim = self.name in _one_dim_output_tasks
        dataset_metric = SkMetric(
            self.metric_func,
            name=self.metric_name,
            one_dim=one_dim,
            greater_is_better=self.greater_is_better,
        )

        return dataset_metric

    @staticmethod
    def _check_loss_from_params(loss_name, loss_params):
        if loss_name in _valid_loss_args:
            required_params = set(_valid_loss_args[loss_name])
        else:
            required_params = set()
        given_params = set(loss_params)
        extra_params = given_params - required_params
        assert len(extra_params) == 0, "For loss {0} given extra params {1}".format(loss_name, extra_params)
        needed_params = required_params - given_params
        assert len(needed_params) == 0, "For loss {0} required params {1} are not defined".format(
            loss_name, needed_params
        )

    @staticmethod
    def _check_metric_from_params(metric_name, metric_params):
        if metric_name in _valid_metric_args:
            required_params = set(_valid_loss_args[metric_name])
        else:
            required_params = set()
        given_params = set(metric_params)
        extra_params = given_params - required_params
        assert len(extra_params) == 0, "For metric {0} given extra params {1}".format(metric_name, extra_params)
        needed_params = required_params - given_params
        assert len(needed_params) == 0, "For metric {0} required params {1} are not defined".format(
            metric_name, needed_params
        )



"""Generate features for data base structure."""

import logging
import re

from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple


try:
    import featuretools as ft
except:
    import warnings

    warnings.warn("Library 'featuretools' - package isn't installed")


import pandas as pd

from ..dataset.base import LAMLDataset
from ..dataset.roles import NumericRole
from ..ml_algo.boost_lgbm import BoostLGBM
from ..pipelines.features.base import TabularDataFeatures
from ..pipelines.features.lgb_pipeline import LGBSimpleFeatures
from ..pipelines.selection.importance_based import ModelBasedImportanceEstimator
from ..pipelines.selection.permutation_importance_based import (
    NpIterativeFeatureSelector,
)
from ..pipelines.utils import get_columns_by_role
from ..reader.utils import set_sklearn_folds
from ..validation.utils import create_validation_iterator
from .base import LAMLTransformer


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_interesting_values(dataset: LAMLDataset, per_top_categories: float) -> Dict:
    """Get interesting value for featuretools.

    Args:
        dataset: Dataset to search.
        per_top_categories: percent of most frequent categories for feature generation in corresponding slices. If number of unique values is less than 10, then the all values are be used.

    Returns:
        Dictionary of categorical feature names with corresponding unique values.

    """
    unique_values_th = 10

    categorical_features = get_columns_by_role(dataset, "Category")
    cat_interesting_values = dict()
    df = dataset.data

    for feature_name in categorical_features:
        df_stat = df[feature_name].value_counts()
        df_stat_len = len(df_stat)

        if df_stat_len > unique_values_th:
            n_top_categories = round(df_stat_len * per_top_categories * 1e-3)
        else:
            n_top_categories = df_stat_len

        cat_interesting_values[feature_name] = df_stat.index[:n_top_categories].tolist()
    return cat_interesting_values


def renaming_function(column_name) -> str:
    """Renaming for appropriate LGBM format.

    Args:
        column_name: name of column.

    Returns:
        renamed column if initial column contains specific symbols

    """
    return re.sub("[^A-Za-z0-9_()=. ]+", "", column_name)


class FeatureGeneratorTransformer(LAMLTransformer, TabularDataFeatures):
    """Generate features.

    Args:
        seq_params: sequence-related params.
        max_gener_features: maximum generated features.
        max_depth: maximum allowed depth of features.
        agg_primitives: list of aggregation primitives.
        trans_primitives: list of transform primitives.
        interesting_values: categorical values if the form of {table_name: {column: [values]}} for feature generation in corresponding slices.
        generate_interesting_values: whether generate feature in slices of unique categories or not.
        per_top_categories: percent of most frequent categories for feature generation in corresponding slices. If number of unique values is less than 10, then the all values are be used.
        sample_size: size of data to make generated feature selection on it.
        n_jobs: number of processes to run in parallel

    """

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "ft"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        seq_params: Dict,
        max_gener_features: int = -1,
        max_depth: int = 2,
        agg_primitives: Optional[Tuple[str]] = (
            "entropy",
            "count",
            "mean",
            "std",
            "median",
            "max",
            "sum",
            "num_unique",
            "min",
            "percent_true",
        ),
        trans_primitives: Optional[Tuple[str]] = (
            "hour",
            "month",
            "weekday",
            "is_weekend",
            "day",
            "time_since_previous",
            "week",
            "age",
            "time_since",
        ),
        interesting_values: Dict[str, Dict[str, List[Any]]] = None,
        generate_interesting_values: bool = False,
        per_top_categories: float = 0.25,
        sample_size: int = None,
        n_jobs: int = 1,
    ):

        super().__init__()

        self.seq_params = deepcopy(seq_params)
        self.max_gener_features = max_gener_features
        self.agg_primitives = agg_primitives
        self.trans_primitives = trans_primitives
        self.interesting_values = interesting_values
        self.max_depth = max_depth
        self.generate_interesting_values = generate_interesting_values
        self.per_top_categories = per_top_categories
        self.sample_size = sample_size
        self.n_jobs = n_jobs

        # Make an entityset
        self.es = ft.EntitySet(id="star_schema")
        self.main_id = "plain_unique_index"

        # Initialize feature selector
        self.feats_imp = NpIterativeFeatureSelector(
            feature_pipeline=LGBSimpleFeatures(),
            ml_algo=BoostLGBM(),
            imp_estimator=ModelBasedImportanceEstimator(),
            fit_on_holdout=False,
            max_features_cnt_in_result=None,
        )

        self.seq_table_names = sorted(list(self.seq_params.keys()))

        self._ignore_columns = None
        self._ft_features = None
        self._features = None
        self.selected_features = None
        self.feature_dict = None

    def _create_entities(self, dataset: LAMLDataset):
        """Create entities."""
        # An entity of the main table
        self.es = self.es.add_dataframe(
            dataframe_name="plain",
            dataframe=dataset.data,
            index=self.main_id,  # unique index
            make_index=True,
        )

        # Define unique columns of ids in plain
        plain_unique_ids = list()
        for seq_table_name in self.seq_table_names:
            if self.seq_params[seq_table_name]["scheme"]["to"] == "plain":
                plain_id = self.seq_params[seq_table_name]["scheme"]["to_id"]
                if plain_id not in plain_unique_ids:
                    plain_unique_ids.append(plain_id)
                    self.es.normalize_dataframe(
                        new_dataframe_name=f"plain_{plain_id}",
                        base_dataframe_name="plain",
                        index=plain_id,
                    )

        # Define new tables' names
        # If in the main table contains two keys, there will be two additional tables with the keys and key to connect with major table.
        plain_i_id = dict()
        for unique_id in plain_unique_ids:
            plain_i_id[unique_id] = f"plain_{unique_id}"
        # TODO: check unused variable 'plain_i_id_inv'
        # plain_i_id_inv = dict((v, k) for k, v in plain_i_id.items())

        # Update defined params for scheme
        for seq_table_name in self.seq_table_names:
            if self.seq_params[seq_table_name]["scheme"]["to"] == "plain":
                self.seq_params[seq_table_name]["scheme"]["to"] = plain_i_id[
                    self.seq_params[seq_table_name]["scheme"]["to_id"]
                ]

        # Create entities for seq tables
        for seq_table_name in sorted(list(self.seq_params.keys())):
            self.es = self.es.add_dataframe(
                dataframe_name=seq_table_name,
                dataframe=dataset.seq_data[seq_table_name].data,
                index=seq_table_name + "_unique_index",
                make_index=True,
            )

    def _add_relationships(self):
        """Add in the defined relationships."""
        for seq_table_name in sorted(list(self.seq_params.keys())):
            scheme = self.seq_params[seq_table_name]["scheme"]
            self.es.add_relationships([(scheme["to"], scheme["to_id"], seq_table_name, scheme["from_id"])])

    def _set_interesting_values(self):
        """Add interesting values if any."""
        for seq_table_name in self.seq_table_names and sorted(list(self.interesting_values.keys())):
            self.es.add_interesting_values(
                dataframe_name=seq_table_name, values=self.interesting_values[seq_table_name]
            )

    def _get_new_feature_names(self):
        """DFS with specified primitives."""
        return ft.dfs(
            entityset=self.es,
            target_dataframe_name="plain",
            agg_primitives=self.agg_primitives,
            where_primitives=self.agg_primitives,
            trans_primitives=self.trans_primitives,
            max_depth=self.max_depth,
            max_features=self.max_gener_features,
            features_only=True,
            ignore_columns={"plain": self._ignore_columns},
        )

    def fit(self, dataset: LAMLDataset) -> "FeatureGeneratorTransformer":
        """Extract nan flags.

        Args:
            dataset: Dataset to make transform.

        Returns:
            self.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)
        # set transformer features

        # Create entities
        self._create_entities(dataset)

        # Fill dict with new interesting values
        if self.generate_interesting_values:

            new_interesting_values = dict()
            for seq_table_name in self.seq_table_names:
                seq_table = dataset.seq_data[seq_table_name]
                new_interesting_values[seq_table_name] = get_interesting_values(seq_table, self.per_top_categories)

            if self.interesting_values is None:
                self.interesting_values = dict()

            self.interesting_values.update(new_interesting_values)
            logger.info3("Interesting values have been generated")

        # Add interesting values
        if bool(self.interesting_values):
            self._set_interesting_values()
            logger.info3("Interesting values have been added to the entityset")

        # Add relationships
        self._add_relationships()
        logger.info3("Relationships have been added to the entityset")

        # DFS with specified primitives
        self._ignore_columns = dataset.data.columns.tolist()
        self._ft_features = self._get_new_feature_names()

        self._features = [
            self._fname_prefix + "__" + renaming_function(feature_name.get_name()) for feature_name in self._ft_features
        ]
        self.feature_dict = dict(zip(self._features, self._ft_features))

        logger.info3(f"{len(self._features)} are going to be generated")
        return self

    def _select(self, dataset: LAMLDataset) -> List[str]:

        if self.sample_size is None:
            self.sample_size = dataset.shape[0]

        df_sample = dataset.data.sample(self.sample_size, random_state=0)

        self.es.replace_dataframe("plain", df_sample)

        fm_sample = ft.calculate_feature_matrix(features=self._ft_features, entityset=self.es, n_jobs=self.n_jobs)

        fm_sample = fm_sample.set_index(df_sample.index)
        fm_sample = fm_sample.rename(columns=dict(zip(fm_sample.columns, self._features)))

        fm_roles_dict = dict(zip(self._features, [NumericRole()] * len(self._features)))

        target_sample = dataset.target.iloc[df_sample.index]
        folds = set_sklearn_folds(task=dataset.task, target=target_sample.values, cv=20)

        dataset_sample = dataset.empty()
        dataset_sample._initialize(task=dataset.task, folds=pd.Series(folds), target=target_sample)
        dataset_sample.set_data(fm_sample, self._features, fm_roles_dict)

        dataset_sample_iter = create_validation_iterator(dataset_sample)

        self.feats_imp.feature_group_size = max(1, int(len(self._features) * 0.15))  # ~15% of the generated features

        self.feats_imp.fit(dataset_sample_iter)

        logger.info3("Selection completed")
        logger.info3(f"{len(self.feats_imp.selected_features)} features have been selected from generated")

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Generate new columns from inrintut dataset.

        Args:
            dataset: Dataset to make transform.

        Returns:
            LAMLDataset with new features.

        """
        # checks here
        super().transform(dataset)

        dataset.data["plain_unique_index"] = dataset.data.index

        if self.feats_imp.is_fitted is False:
            self._select(dataset)
            self.selected_features = self.feats_imp.selected_features

        fm_roles_dict = dict(zip(self.selected_features, [NumericRole()] * len(self.selected_features)))

        self.es.replace_dataframe("plain", dataset.data)

        fm = ft.calculate_feature_matrix(
            features=[self.feature_dict.get(feature) for feature in self.selected_features],
            entityset=self.es,
            n_jobs=self.n_jobs,
        )

        fm = fm.rename(columns=dict(zip(fm.columns, self.selected_features)))

        dataset_out = dataset.empty()
        dataset_out.set_data(fm, self.selected_features, fm_roles_dict)

        logger.info3("Feature generation is completed")
        return dataset_out



"""Uplift metrics."""

import abc

from typing import Tuple

import numpy as np

from sklearn.metrics import auc
from sklearn.utils.multiclass import type_of_target


_available_uplift_modes = ("qini", "cum_gain", "adj_qini")


class ConstPredictError(Exception):
    """Exception: constant prediction."""

    pass


class TUpliftMetric(metaclass=abc.ABCMeta):
    """Base Uplift Metric."""

    @abc.abstractclassmethod
    def __call__(self, y_true: np.ndarray, uplift_pred: np.ndarray, treatment: np.ndarray) -> float:
        """Calculate metric."""
        pass


def perfect_uplift_curve(y_true: np.ndarray, treatment: np.ndarray) -> np.ndarray:
    """Calculate perfect curve.

    Method return curve's coordinates if the model is a perfect.
    Perfect model ranking:
        If type if 'y_true' is 'binary':
            1) Treatment = 1, Target = 1
            2) Treatment = 0, Target = 0
            3) Treatment = 1, Target = 0
            4) Treatment = 0, Target = 1

        If type if 'y_true' is 'continuous':
            Not implemented

    Args:
        y_true: Target values
        treatment: Treatment column

    Returns:
        perfect curve

    """
    if type_of_target(y_true) == "binary":
        perfect_control_score = (treatment == 0).astype(int) * (2 * (y_true != 1).astype(int) - 1)
        perfect_treatment_score = (treatment == 1).astype(int) * 2 * (y_true == 1).astype(int)
        perfect_uplift = perfect_treatment_score + perfect_control_score
    elif type_of_target(y_true) == "continuous":
        raise NotImplementedError("Can't calculate perfect curve for continuous target")
    else:
        raise RuntimeError("Only 'binary' and 'continuous' targets are available")

    return perfect_uplift


def _get_uplift_curve(
    y_treatment: np.ndarray, y_control: np.ndarray, n_treatment: np.ndarray, n_control: np.ndarray, mode: str
):
    """Calculate uplift curve.

    Args:
        y_treatment: Cumulative number of target in treatment group
        y_control: Cumulative number of target in control group
        n_treatment: Cumulative number of treatment group
        n_control: Cumulative number of treatment group
        mode: Name of available metrics

    Returns:
        curve for current mode

    """
    assert mode in _available_uplift_modes, "Mode isn't available"

    if mode == "qini":
        curve_values = y_treatment / n_treatment[-1] - y_control / n_control[-1]
    elif mode == "cum_gain":
        treatment_target_rate = np.nan_to_num(y_treatment / n_treatment, 0.0)
        control_target_rate = np.nan_to_num(y_control / n_control, 0.0)
        curve_values = treatment_target_rate - control_target_rate
        n_join = n_treatment + n_control
        curve_values = curve_values * n_join / n_join[-1]
    elif mode == "adj_qini":
        normed_factor = np.nan_to_num(n_treatment / n_control, 0.0)
        normed_y_control = y_control * normed_factor
        curve_values = (y_treatment - normed_y_control) / n_treatment[-1]

    return curve_values


def calculate_graphic_uplift_curve(
    y_true: np.ndarray, uplift_pred: np.ndarray, treatment: np.ndarray, mode: str = "adj_qini"
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate uplift curve.

    Args:
        y_true: Target values
        uplift_pred: Prediction of models
        treatment: Treatment column
        mode: Name of available metrics

    Returns:
        xs, ys - curve's coordinates

    """
    # assert not np.all(uplift_pred == uplift_pred[0]), "Can't calculate uplift curve for constant predicts"
    if np.all(uplift_pred == uplift_pred[0]):
        raise ConstPredictError("Can't calculate uplift curve for constant predicts")

    sorted_indexes = np.argsort(uplift_pred)[::-1]
    y_true, uplift_pred, treatment = (
        y_true[sorted_indexes],
        uplift_pred[sorted_indexes],
        treatment[sorted_indexes],
    )

    indexes = np.where(np.diff(uplift_pred))[0]
    indexes = np.insert(indexes, indexes.size, uplift_pred.shape[0] - 1)

    n_treatment_samples_cs = np.cumsum(treatment)[indexes].astype(np.int64)
    n_join_samples_cs = indexes + 1
    n_control_samples_cs = n_join_samples_cs - n_treatment_samples_cs

    y_true_control, y_true_treatment = y_true.copy(), y_true.copy()

    y_true_control[treatment == 1] = 0
    y_true_treatment[treatment == 0] = 0

    y_true_control_cs = np.cumsum(y_true_control)[indexes]
    y_true_treatment_cs = np.cumsum(y_true_treatment)[indexes]

    curve_values = _get_uplift_curve(
        y_true_treatment_cs,
        y_true_control_cs,
        n_treatment_samples_cs,
        n_control_samples_cs,
        mode,
    )

    n_join_samples = np.insert(n_join_samples_cs, 0, 0)
    curve_values = np.insert(curve_values, 0, 0)
    rate_join_samples = n_join_samples / n_join_samples[-1]

    return rate_join_samples, curve_values


def calculate_uplift_auc(
    y_true: np.ndarray, uplift_pred: np.ndarray, treatment: np.ndarray, mode: str = "adj_qini", normed: bool = False
):
    """Calculate area under uplift curve.

    Args:
        y_true: Target values
        uplift_pred: Prediction of meta model
        treatment: Treatment column
        mode: Name of available metrics
        normed: Normed AUC: (AUC - MIN_AUC) / (MAX_AUC - MIN_AUC)

    Returns:
        auc_score: Area under model uplift curve

    """
    xs, ys = calculate_graphic_uplift_curve(y_true, uplift_pred, treatment, mode)

    uplift_auc = auc(xs, ys)

    if normed:
        (
            min_auc,
            max_auc,
        ) = calculate_min_max_uplift_auc(y_true, treatment, mode)

        uplift_auc = (uplift_auc - min_auc) / (max_auc - min_auc)

    return uplift_auc


def calculate_min_max_uplift_auc(y_true: np.ndarray, treatment: np.ndarray, mode: str = "adj_qini"):
    """Calculate AUC uplift curve for `base` and `perfect` models.

    Args:
        y_true: Target values
        treatment: Treatment column
        mode: Name of available metrics

    Returns:
        auc_base: Area under `base`.
        auc_perfect: Area under `perfect` model curve

    """
    diff_target_rate = y_true[treatment == 1].mean() - y_true[treatment == 0].mean()
    xs_base, ys_base = np.array([0, 1]), np.array([0, diff_target_rate])

    perfect_uplift = perfect_uplift_curve(y_true, treatment)
    xs_perfect, ys_perfect = calculate_graphic_uplift_curve(y_true, perfect_uplift, treatment, mode)

    auc_base = auc(xs_base, ys_base)
    auc_perfect = auc(xs_perfect, ys_perfect)

    return auc_base, auc_perfect


def calculate_uplift_at_top(y_true: np.ndarray, uplift_pred: np.ndarray, treatment: np.ndarray, top: float = 30):
    """Calculate Uplift metric at TOP.

    Calculate uplift metric at top

    Args:
        y_true: Target values
        uplift_pred: Prediction of meta model
        treatment: Treatment column
        top: Rate, value between (0, 100]

    Returns:
        score: Score

    """
    # assert not np.all(uplift_pred == uplift_pred[0]), "Can't calculate for constant predicts."

    uplift_percentile = np.percentile(uplift_pred, 100 - top)
    mask_top = uplift_pred > uplift_percentile

    control_true_top = y_true[(treatment == 0) & mask_top].sum()
    treatment_true_top = y_true[(treatment == 1) & mask_top].sum()

    n_control_samples = (treatment[mask_top] == 0).sum()
    n_treatment_samples = (treatment[mask_top] == 1).sum()

    mean_control_value = control_true_top / n_control_samples if n_control_samples > 0 else 0.0
    mean_treatment_value = treatment_true_top / n_treatment_samples if n_treatment_samples > 0 else 0.0

    score = mean_treatment_value - mean_control_value

    return score


def calculate_total_score(y_true: np.ndarray, uplift_pred: np.ndarray, treatment: np.ndarray, top: float = 30):
    """Calculate total target.

    Args:
        y_true: Target values
        uplift_pred: Prediction of meta model
        treatment: Treatment column
        top: Rate, value between (0, 100]


    Returns:
        score: Score

    """
    uplift_percentile = np.percentile(uplift_pred, 100 - top)
    mask_top = uplift_pred > uplift_percentile
    mask_treatment = treatment == 1
    treatment_rate = mask_treatment.mean()

    control_true_top = y_true[(~mask_treatment) & (~mask_top)].mean()
    treatment_true_top = y_true[mask_treatment & mask_top].mean()

    score = control_true_top * (1 - treatment_rate) + treatment_true_top * treatment_rate

    return score



"""Pipeline for neural network models."""

from typing import Optional
from typing import Union

import numpy as np

from ...dataset.np_pd_dataset import NumpyDataset
from ...dataset.np_pd_dataset import PandasDataset
from ...dataset.roles import NumericRole
from ...transformers.base import ChangeRoles
from ...transformers.base import ColumnsSelector
from ...transformers.base import ConvertDataset
from ...transformers.base import LAMLTransformer
from ...transformers.base import SequentialTransformer
from ...transformers.base import UnionTransformer
from ...transformers.categorical import LabelEncoder
from ...transformers.datetime import TimeToNum
from ...transformers.numeric import FillInf
from ...transformers.numeric import FillnaMean
from ...transformers.numeric import QuantileTransformer
from ...transformers.numeric import StandardScaler
from ..utils import get_columns_by_role
from .base import FeaturesPipeline
from .base import TabularDataFeatures


NumpyOrPandas = Union[PandasDataset, NumpyDataset]


class TorchSimpleFeatures(FeaturesPipeline, TabularDataFeatures):
    """Creates simple pipeline for neural network models."""

    def __init__(
        self,
        use_te: bool = False,
        top_intersections: int = 5,
        max_bin_count: int = 10,
        max_intersection_depth: int = 3,
        te_subsample: Optional[Union[int, float]] = None,
        sparse_ohe: Union[str, bool] = "auto",
        auto_unique_co: int = 50,
        output_categories: bool = True,
        multiclass_te_co: int = 3,
        use_qnt: bool = True,
        n_quantiles: Optional[int] = None,
        subsample: int = 10 ** 9,
        output_distribution: str = "normal",
        noise: float = 1e-3,
        qnt_factor: int = 30,
        **kwargs
    ):
        """TorchSimpleFeatures.

        Args:
            use_qnt: Use quantile transformer for numerical columns.
            n_quantiles: Number of quantiles to be computed.
            subsample: Maximum number of samples used to estimate the quantiles for computational efficiency.
            output_distribution: Marginal distribution for the transformed data. The choices are 'uniform' or 'normal'.
            noise: Add noise with certain std to dataset before quantile transformation to make data more smooth.
            qnt_factor: If number of quantiles is none then it equals dataset size / factor
            use_te: Use target encoding for categorical columns.
            top_intersections: Max number of categories to generate intersections.
            max_bin_count: Max number of bins for cat columns.
            max_intersection_depth: Max depth of cat intersection.
            te_subsample: Subsample to calc data statistics
            sparse_ohe: Should we output sparse if ohe encoding was used during cat handling.
            auto_unique_co: Switch to target encoding if high cardinality.
            output_categories: Output encoded categories or embed idxs.
            multiclass_te_co: Cutoff if use target encoding in cat handling on multiclass task if number of classes is high.
            kwargs: Other params.
        """
        super().__init__(
            multiclass_te=False,
            top_intersections=top_intersections,
            max_intersection_depth=max_intersection_depth,
            subsample=te_subsample,
            auto_unique_co=auto_unique_co,
            output_categories=output_categories,
            ascending_by_cardinality=True,
            max_bin_count=max_bin_count,
            sparse_ohe=sparse_ohe,
            multiclass_te_co=multiclass_te_co,
            **kwargs
        )
        self.use_qnt = use_qnt
        self.n_quantiles = n_quantiles
        self.subsample = subsample
        self.output_distribution = output_distribution
        self.noise = noise
        self.qnt_factor = qnt_factor
        self.use_te = use_te

    def create_pipeline(self, train: NumpyOrPandas) -> LAMLTransformer:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        transformers_list = []
        cat_cols = get_columns_by_role(train, "Category")
        freq_cols = get_columns_by_role(train, "Category", encoding_type="freq")
        other_cols = sorted(list(set(cat_cols) - set(freq_cols)))

        transformers_list.append(self.get_freq_encoding(train, freq_cols))

        # process categories
        if len(other_cols) > 0:
            cat_processing = SequentialTransformer([ColumnsSelector(keys=other_cols), LabelEncoder()])

            if self.use_te:
                target_encoder = self.get_target_encoder(train)

                # get target encoded categories
                te_part = self.get_categorical_raw(train, other_cols)
                if te_part is not None and target_encoder is not None:
                    transformers_list.append(SequentialTransformer([te_part, target_encoder()]))

                # get intersection of top categories
                intersections = self.get_categorical_intersections(train)
                if intersections is not None and target_encoder is not None:
                    transformers_list.append(SequentialTransformer([intersections, target_encoder()]))
            else:
                transformers_list.append(cat_processing)

        # process datetimes
        datetimes = get_columns_by_role(train, "Datetime")
        if len(datetimes) > 0:
            dt_processing = SequentialTransformer([ColumnsSelector(keys=datetimes), TimeToNum(), FillnaMean()])
            transformers_list.append(dt_processing)

        # process numbers
        numerics = get_columns_by_role(train, "Numeric")
        if len(numerics) > 0:
            num_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=numerics),
                    FillInf(),
                    FillnaMean(),
                    QuantileTransformer(
                        n_quantiles=self.n_quantiles,
                        subsample=self.subsample,
                        output_distribution=self.output_distribution,
                        noise=self.noise,
                        qnt_factor=self.qnt_factor,
                    )
                    if self.use_qnt
                    else StandardScaler(),
                    ConvertDataset(dataset_type=NumpyDataset),
                    ChangeRoles(NumericRole(np.float32)),
                ]
            )
            transformers_list.append(num_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all



"""Lime explanation method for texts."""

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from sklearn.linear_model import Ridge
from sklearn.linear_model import lars_path
from sklearn.metrics import pairwise_distances
from sklearn.utils import check_random_state

from ...pipelines.features.text_pipeline import _tokenizer_by_lang
from .utils import IndexedString
from .utils import draw_html


class TextExplanation:
    """Explanation of object for textual data.

    Args:
        index_string: Pertrubing string object.
        task_name: Task name. Can be
            one of ['binary', 'multiclass', 'reg'].
        prediction:
        class_names: List of class names.
        random_state: Random seed for perturbation generation.
        draw_prediction:

    """

    def __init__(
        self,
        index_string: IndexedString,
        task_name: str,
        prediction: np.ndarray,
        class_names: Optional[List[Any]] = None,
        random_state=None,
        draw_prediction: bool = False,
    ):
        self.idx_str = index_string
        assert task_name in ["binary", "multiclass", "reg"]
        self.task_name = task_name
        self.prediction = prediction
        self.class_names = class_names
        self.draw_prediction = draw_prediction

        if task_name == "reg":
            self.default_label = 0
        else:
            self.default_label = 1

        self.random_state = check_random_state(random_state)
        self.instance = {}

    def as_list(self, label: Optional[int] = None) -> List[Tuple[int, float]]:
        """Get feature weights as list.

        Args:
            label: Explaing label. Not necessary
                for regression. By default, for
                regression 0 will be used, and 1 for
                other task types.

        Returns:
            List in format (token_id, weight).

        """
        label = self._label(label)
        ans = self.instance[label]["feature_weights"]
        ans = [(x[0], float(x[1])) for x in ans]

        return ans

    def as_features(
        self,
        label: Optional[int] = None,
        add_not_rel: bool = False,
        normalize: bool = False,
    ) -> List[Tuple[str, float]]:
        """Get feature weights as list with feature names.

        Args:
            label: Explaing label. Not necessary
                for regression. By default, for
                regression 0 will be used, and 1 for
                other task types.
            add_not_rel: The not relevalt tokens will be
                added in explanation with zero feature weights.
                Using this flag it is convenient to visualize
                the predictions.
            normalize: Normalization by the maximum
                absolute value of the importance of the feature.

        Returns:
            List in format (token, weight).

        """
        label = self._label(label)
        fw = self.instance[label]["feature_weights"]
        norm_const = 1.0
        if normalize:
            norm_const = 1 / abs(fw[0][1])
        fw = dict(fw)
        if add_not_rel:
            fw = dict(self.instance[label]["feature_weights"])
            weights = np.zeros_like(self.idx_str.as_np_, dtype=np.float32)
            for k, v in fw.items():
                weights[self.idx_str.pos[k]] = v

            ans = [(k, float(w) * norm_const) for k, w in zip(self.idx_str.as_np_, weights)]
        else:
            ans = [(self.idx_str.word(k), float(v) * norm_const) for k, v in fw.items()]

        return ans

    def as_map(self, label: Optional[int] = None) -> Dict[str, float]:
        """Get feature weights as list with features.

        Args:
            label: Explaing label. Not necessary
                for regression. By default, for
                regression 0 will be used, and 1 for
                other task types.

        Returns:
            Dictonary of tokens and it's weights in format
            ({token_id}_{position}, feature_weight).

        """
        label = self._label(label)
        return {f"{k}_{i}": v for i, (k, v) in enumerate(self.instance[label]["feature_weights"])}

    def as_html(self, label: Optional[int] = None) -> str:
        """Generates inline HTML with colors.

        Args:
            label: Explaing label. Not necessary
                for regression. By default, for
                regression 0 will be used, and 1 for
                other task types.

        Returns:
            HTML code.

        """
        label = self._label(label)
        weight_string = self.as_features(label, add_not_rel=True, normalize=False)
        prediction = self.prediction[label]
        if not self.draw_prediction:
            prediction = None

        if self.task_name == "reg":
            return draw_html(weight_string, self.task_name, prediction=prediction)
        elif self.task_name == "binary":
            return draw_html(
                weight_string,
                self.task_name,
                prediction=prediction,
                grad_line=True,
                grad_positive_label=str(self.class_names[label]),
                grad_negative_label=str(self.class_names[1 - label]),
            )
        elif self.task_name == "multiclass":
            return draw_html(
                weight_string,
                self.task_name,
                prediction=prediction,
                grad_line=True,
                grad_positive_label=str(self.class_names[label]),
            )

    def visualize_in_notebook(self, label: Optional[int] = None):
        """Visualization of interpretation in IPython notebook.

        Args:
            label: Explaing label. Not necessary
                for regression. By default, for
                regression 0 will be used, and 1 for
                other task types.

        """
        from IPython.display import HTML
        from IPython.display import display_html

        label = self._label(label)
        raw_html = self.as_html(label)
        display_html(HTML(raw_html))

    def _label(self, label: Union[None, int]) -> int:
        if label is None or self.task_name == "reg":
            label = self.default_label

        return label

    def get_label(self, name):
        """Return label of class."""
        return self.class_names.index(name)


class LimeTextExplainer:
    """Instance-wise textual explanation.

    Method working as follows:
    1. Tokenize perturbed text-column.
    2. Create dataset by perturbing text column.
    3. Select features (Optional).
    4. Fit explainable model (distil_model). (For multiclass
    one-versus-all will be used)
    5. Get the explanation from distil_model.

    Note:
        More info: `"Why Should I Trust You?":
        Explaining the Predictions of Any Classifier
        Ribeiro et al. <https://arxiv.org/abs/1602.04938>`_

    Note:
        Basic usage of explaier.

        >>> task = Task('reg')
        >>> automl = TabularNLPAutoML(task=task,
        >>>     timeout=600, gpu_ids = '0',
        >>>     general_params = {'nested_cv': False, 'use_algos': [['nn']]},
        >>>     text_params = {'lang': 'ru'})
        >>> automl.fit_predict(train, roles=roles)
        >>> lime = LimeTextExplainer(automl)
        >>> explanation = lime.explain_instance(train.iloc[0], perturb_column='message')
        >>> explanation.visualize_in_notebook()

    Args:
        automl: Automl object.
        kernel: Callable object with parameter `kernel_width`.
            By default, the squared-exponential kernel will be used.
        kernel_width: Kernel width.
        feature_selection: Feature selection type. For now,
            'none', 'lasso' are availiable.
        force_order: Whether to follow the word order.
        model_regressor: Model distilator. By default,
            Ridge regression will be used.
        distance_metric: Distance type between binary vectors.
        random_state: Random seed used,
            for sampling perturbation of text column.
        draw_prediction:


    """

    def __init__(
        self,
        automl,
        kernel: Optional[Callable] = None,
        kernel_width: float = 25.0,
        feature_selection: str = "none",
        force_order: bool = False,
        model_regressor: Any = None,
        distance_metric: str = "cosine",
        random_state: Union[int, np.random.RandomState] = 0,
        draw_prediction: bool = False,
    ):
        self.automl = automl
        self.task_name = automl.reader.task.name

        assert self.task_name in ["binary", "multiclass", "reg"]

        self.roles = automl.reader.roles
        self.kernel_width = kernel_width
        if kernel is None:

            def kernel(d, kernel_width):
                return np.sqrt(np.exp(-(d ** 2) / kernel_width ** 2))

        assert callable(kernel)

        self.kernel_fn = partial(kernel, kernel_width=kernel_width)
        self.random_state = check_random_state(random_state)

        if model_regressor is None:
            model_regressor = Ridge(alpha=1, fit_intercept=True, random_state=self.random_state)

        self.distil_model = model_regressor

        # todo: forward selection and higher weights
        # and l2x
        # and auto-mode

        assert feature_selection in ["none", "lasso"]

        self.feature_selection = feature_selection
        self.force_order = force_order

        lang = automl.text_params["lang"]
        self.tokenizer = _tokenizer_by_lang[lang](is_stemmer=False)
        self.distance_metric = distance_metric

        class_names = automl.reader.class_mapping
        if class_names is None:
            if self.task_name == "reg":
                class_names = [0]
            else:
                class_names = np.arange(automl.reader._n_classes)
        else:
            class_names = list(class_names.values())

        self.class_names = class_names
        self.draw_prediction = draw_prediction

    def explain_instance(
        self,
        data: pd.Series,
        perturb_column: str,
        labels: Optional[Iterable] = None,
        n_features: int = 10,
        n_samples: int = 5000,
    ) -> "TextExplanation":
        """Create instance of TextExplanation.

        Args:
            data: Data sample to explain.
            perturb_column: Column that will be perturbed.
            labels: Target variable values to be interpreted.
            n_features: If a feature selector was specified,
                then this number means the maximum number
                of features that will be used in the distilled model.
            n_samples: Number of sampled instances by perturbing text column.

        Returns:
            TextExplanation object.

        """
        assert self.roles[perturb_column].name == "Text", "Column is not text column"
        assert n_samples > 1, "Number of generated samples must be > 0"

        if labels is None:
            labels = (0,) if self.task_name == "reg" else (1,)

        data, y, dst, expl = self._get_perturb_dataset(data, perturb_column, n_samples)

        for label in labels:
            expl.instance[label] = self._explain_dataset(data, y, dst, label, n_features)

        return expl

    def _get_perturb_dataset(self, data, perturb_column: str, n_samples: int):
        text = data[perturb_column]
        idx_str = IndexedString(text, self.tokenizer, self.force_order)
        n_words = idx_str.n_words
        samples = self.random_state.randint(1, n_words + 1, n_samples - 1)
        raw_dataset = [data.copy()]
        dataset = np.ones((n_samples, n_words))

        for i, size in enumerate(samples, start=1):
            off_tokens = self.random_state.choice(range(n_words), size)
            data_ = data.copy()
            p_text = idx_str.inverse_removing(off_tokens)
            data_[perturb_column] = p_text
            raw_dataset.append(data_)
            dataset[i, off_tokens] = 0

        raw_dataset = pd.DataFrame(raw_dataset)

        pred = self.automl.predict(raw_dataset).data
        if self.task_name == "binary":
            pred = np.concatenate([1 - pred, pred], axis=1)

        distance = pairwise_distances(dataset, dataset[0].reshape(1, -1), metric=self.distance_metric).ravel()

        expl = TextExplanation(idx_str, self.task_name, pred[0], self.class_names, self.random_state)

        return dataset, pred, distance * 100, expl

    def _explain_dataset(
        self,
        data: pd.DataFrame,
        y: np.array,
        dst: np.array,
        label: int,
        n_features: int,
    ) -> Dict[str, Union[float, np.array]]:
        weights = self.kernel_fn(dst)
        y = y[:, label]
        features = self._feature_selection(data, y, weights, n_features, mode=self.feature_selection)
        model = self.distil_model
        model.fit(data[:, features], y, sample_weight=weights)
        score = model.score(data[:, features], y, sample_weight=weights)

        pred = model.predict(data[0, features].reshape(1, -1))
        feature_weights = list(sorted(zip(features, model.coef_), key=lambda x: np.abs(x[1]), reverse=True))
        res = {
            "bias": model.intercept_,
            "feature_weights": feature_weights,
            "score": score,
            "pred": pred,
        }

        return res

    def _feature_selection(
        self,
        data: pd.DataFrame,
        y: np.array,
        weights: np.array,
        n_features: int,
        mode: str = "none",
    ) -> List[int]:
        if mode == "none":
            return np.arange(data.shape[1])
        if mode == "lasso":
            weighted_data = (data - np.average(data, axis=0, weights=weights)) * np.sqrt(weights[:, np.newaxis])
            weighted_y = (y - np.average(y, weights=weights)) * np.sqrt(weights)

            features = np.arange(weighted_data.shape[1])
            _, _, coefs = lars_path(weighted_data, weighted_y, method="lasso", verbose=False)

            for i in range(len(coefs.T) - 1, 0, -1):
                features = coefs.T[i].nonzero()[0]
                if len(features) <= n_features:
                    break

            return features
